from pathlib import Path
from io import BytesIO
from typing import Optional, Sequence

import numpy as np

try:
    from pygraphviz import AGraph  # type: ignore
except ImportError:

    class AGraph:
        def _raise(self):
            raise AttributeError(
                "`pygraphviz` not installed! Please follow the documentation "
                "to install `graphviz`, then `pip instal pipelime[draw]`."
            )

        def __init__(self, *args, **kwargs):
            self._raise()

        def __getattribute__(self, name):
            self._raise()


from pipelime.piper.drawing.base import NodesGraphDrawer
from pipelime.piper.graph import (
    DAGNodesGraph,
    GraphNode,
    GraphNodeData,
    GraphNodeOperation,
)


class GraphvizNodesGraphDrawer(NodesGraphDrawer):
    def _is_node_file(self, node: GraphNodeData) -> bool:
        """Returns True if the node path is (or could be) a file

        Args:
            node (GraphNodeData):  input node

        Returns:
            bool:  True if the node path is (or could be) a file
        """
        return len(Path(node.path).suffix) > 0

    def _operation_shape(self, node: GraphNodeOperation) -> str:
        """Returns the shape string of an operation node

        Args:
            node (GraphNodeOperation): input node

        Returns:
            str: shape of the node
        """
        return "box3d"

    def _data_shape(self, node: GraphNodeData) -> str:
        """Returns the shape string of a data node

        Args:
            node (GraphNodeData): input node

        Returns:
            str: shape of the node
        """
        return "note" if self._is_node_file(node) else "cylinder"

    def _operation_styles(self, node: GraphNodeOperation) -> dict:
        """Returns the style attributes for a generic Operation node

        Args:
            node (GraphNodeOperation): input node

        Returns:
            dict: style attributes for the node
        """
        return {
            "style": "filled",
            "fillcolor": "#03A9F4",
            "fontcolor": "#fafafa",
        }

    def _data_styles(self, node: GraphNodeData) -> dict:
        """Returns the style attributes for a generic Data node

        Args:
            node (GraphNodeData): input node

        Returns:
            dict: style attributes for the node
        """
        if self._is_node_file(node):
            return {
                "style": "filled",
                "fillcolor": "#fafafa",
                "fontcolor": "#222222",
            }
        else:
            return {
                "style": "filled",
                "fillcolor": "#009688",
                "fontcolor": "#fafafa",
            }

    def _operation_attributes(self, node: GraphNodeOperation) -> dict:
        """Returns the whole attributes kwargs for a generic Operation node

        Args:
            node (GraphNodeOperation): input node

        Returns:
            dict: attributes for the node
        """
        return {
            **(self._operation_styles(node)),
            "shape": self._operation_shape(node),
            "label": self._operation_node_content(node),
            # Easter egg
            "URL": "https://c.tenor.com/2jvzfQQxDFcAAAAC/think-smart.gif",
        }

    def _data_attributes(self, node: GraphNodeData) -> dict:
        """Returns the whole attributes kwargs for a generic Data node

        Args:
            node (GraphNodeData): input node

        Returns:
            dict: attributes for the node
        """
        return {
            **(self._data_styles(node)),
            "shape": self._data_shape(node),
            "label": self._data_node_content(node),
        }

    def _operation_node_content(self, node: GraphNodeOperation) -> str:
        """Returns the textual representation of an Operation node in Graphviz TABLE
        format

        Args:
            node (GraphNodeOperation): input node

        Returns:
            str: textual representation of the node
        """
        content = '<<TABLE BORDER="0" CELLBORDER="0" CELLPADDING="20">'
        content += "<TR>"
        content += f"<TD>{node.command.command_name()}</TD>"
        content += "</TR>"
        content += "</TABLE>>"
        return content

    def _data_node_content(self, node: GraphNodeData) -> str:
        """Returns the textual representation of a Data node in Graphviz TABLE format

        Args:
            node (GraphNodeData): input node

        Returns:
            str: textual representation of the node
        """
        content = '<<TABLE BORDER="0" CELLBORDER="0" CELLPADDING="20">'
        content += "<TR>"
        content += f"<TD>{node.path}</TD>"
        content += "</TR>"
        content += "</TABLE>>"
        return content

    def _add_node_to_agraph(self, agraph: AGraph, node: GraphNode):
        """Adds a node to the graph distinguising between operation and data nodes

        Args:
            agraph (AGraph): graph to add the node to
            node (GraphNode): node to add
        """
        if isinstance(node, GraphNodeOperation):
            agraph.add_node(node, **(self._operation_attributes(node)))
        elif isinstance(node, GraphNodeData):
            agraph.add_node(node, **(self._data_attributes(node)))

    def _data_edge_style(self, node: GraphNodeData, edge_attrs: dict) -> dict:
        """Returns the style attributes for a data edge (an edge starting from a Data
        node and ending in an Operation node)

        Args:
            node (GraphNodeData): input node
            edge_attrs (dict): edge attributes

        Returns:
            dict: style attributes for the edge
        """
        return {
            "color": "#1DE9B6",
            "arrowsize": 2.0,
            "style": "dashed",
        }

    def _operation_edge_style(self, node: GraphNodeOperation, edge_attrs: dict) -> dict:
        """Returns the style attributes for an operation edge (an edge starting from an
        Operation node and ending in a Data node)

        Args:
            node (GraphNodeOperation): input node
            edge_attrs (dict): edge attributes

        Returns:
            dict: style attributes for the edge
        """
        return {
            "color": "#00B0FF",
            "arrowsize": 1.0,
        }

    def _add_edge_op2data(
        self,
        agraph: AGraph,
        n0: GraphNodeOperation,
        n1: GraphNodeData,
        edge_attrs: dict,
    ):
        """Adds an edge from an Operation node to a Data node

        Args:
            agraph (AGraph): graph to add the edge to
            n0 (GraphNodeOperation): starting node
            n1 (GraphNodeData): ending node
            edge_attrs (dict): edge attributes
        """

        edge_name = edge_attrs.get(DAGNodesGraph.GraphAttrs.OUTPUT_PORT, "")
        agraph.add_edge(
            n0,
            n1,
            label=f"{edge_name}",
            **(self._data_edge_style(n1, edge_attrs)),
        )

    def _add_edge_data2op(
        self,
        agraph: AGraph,
        n0: GraphNodeData,
        n1: GraphNodeOperation,
        edge_attrs: dict,
    ):
        """Adds an edge from a Data node to an Operation node

        Args:
            agraph (AGraph): graph to add the edge to
            n0 (GraphNodeData): starting node
            n1 (GraphNodeOperation): ending node
            edge_attrs (dict): edge attributes
        """

        edge_name = edge_attrs.get(DAGNodesGraph.GraphAttrs.INPUT_PORT, "")
        agraph.add_edge(
            n0,
            n1,
            label=f"{edge_name}",
            **(self._operation_edge_style(n1, edge_attrs)),
        )

    def _add_edge_to_agraph(
        self,
        agraph: AGraph,
        n0: GraphNode,
        n1: GraphNode,
        edge_attrs: dict,
    ):
        """Adds an edge to the graph

        Args:
            agraph (AGraph): graph to add the edge to
            n0 (GraphNode): starting node
            n1 (GraphNode): ending node
            edge_attrs (dict): edge attributes

        Raises:
            ValueError: if the edge is not managed
        """

        if isinstance(n0, GraphNodeOperation) and isinstance(n1, GraphNodeData):
            self._add_edge_op2data(agraph, n0, n1, edge_attrs)
        elif isinstance(n0, GraphNodeData) and isinstance(n1, GraphNodeOperation):
            self._add_edge_data2op(agraph, n0, n1, edge_attrs)
        else:
            raise ValueError(f"Unknown edge type: {type(n0)}, {type(n1)}")

    def _build_agraph(self, graph: DAGNodesGraph) -> AGraph:
        """Builds the graphviz graph from a DAGNodesGraph

        Args:
            graph (DAGNodesGraph): input graph

        Returns:
            AGraph: graphviz graph
        """
        agraph = AGraph(directed=True)

        for edge in graph.raw_graph.edges():
            n0: GraphNode = edge[0]
            n1: GraphNode = edge[1]

            self._add_node_to_agraph(agraph, n0)
            self._add_node_to_agraph(agraph, n1)
            self._add_edge_to_agraph(agraph, n0, n1, graph.raw_graph.edges[n0, n1])

        return agraph

    def draw(self, graph: DAGNodesGraph) -> np.ndarray:
        """Draws a graph and returns the image as a numpy array

        Args:
            graph (DAGNodesGraph): graph to draw

        Returns:
            np.ndarray: image as a numpy array
        """
        import imageio

        agraph: AGraph = self._build_agraph(graph)
        agraph.layout("dot")
        bmp_img = agraph.draw(format="bmp", prog="dot")
        return imageio.imread(BytesIO(bmp_img))

    def representation(self, graph: DAGNodesGraph) -> str:
        """Returns a representation of the graph as a DOT string

        Args:
            graph (DAGNodesGraph): graph to represent

        Returns:
            str: DOT string
        """

        agraph = self._build_agraph(graph)

        return agraph.string()

    def exportable_formats(self) -> Sequence[str]:
        """Returns the exportable formats. They are mixed formats (png, pdf, ...) and
        the dot format.

        Returns:
            Sequence[str]: exportable formats
        """
        return ["png", "dot", "pdf", "svg"]

    def export(self, graph: DAGNodesGraph, filename: str, format: Optional[str] = None):
        """Exports the graph to a file

        Args:
            graph (DAGNodesGraph): graph to export
            filename (str): filename
            format (str): format
        """
        super().export(graph, filename, format)

        if format is None:
            format = Path(filename).suffix[1:]

        agraph = self._build_agraph(graph)
        if format in ["dot"]:
            agraph.write(filename)
        else:
            agraph.layout("dot")
            agraph.draw(filename)



from pathlib import Path
from typing import Any

import numpy as np
import pytest
from pipelime.choixe.ast.nodes import DictNode, ListNode, Node, LiteralNode
from pipelime.choixe.visitors import decode
from deepdiff import DeepDiff
from pydantic import BaseModel


class Person(BaseModel):
    id_: str
    age: int


class Cat(BaseModel):
    age: int
    weight: float
    name: str
    owner: Person


@pytest.mark.parametrize(
    ["node", "expected"],
    [
        [
            DictNode(
                nodes={LiteralNode(data="foo"): LiteralNode(data=np.zeros((2, 3, 2)))}
            ),
            {
                "foo": [
                    [[0.0, 0.0], [0.0, 0.0], [0.0, 0.0]],
                    [[0.0, 0.0], [0.0, 0.0], [0.0, 0.0]],
                ]
            },
        ],
        [
            ListNode(LiteralNode(data=np.zeros((2, 3, 2)))),
            [
                [
                    [[0.0, 0.0], [0.0, 0.0], [0.0, 0.0]],
                    [[0.0, 0.0], [0.0, 0.0], [0.0, 0.0]],
                ]
            ],
        ],
        [LiteralNode(data=np.uint8(24)), 24],
        [LiteralNode(data=np.float64(0.125)), 0.125],
        [
            LiteralNode(
                data=Cat(
                    age=10,
                    weight=5.23,
                    name="Oliver",
                    owner=Person(id_="OCJ123j", age=32),
                )
            ),
            {
                "$model": "tests.pipelime.choixe.visitors.test_decoder.Cat",
                "$args": {
                    "age": 10,
                    "weight": 5.23,
                    "name": "Oliver",
                    "owner": {
                        "id_": "OCJ123j",
                        "age": 32,
                    },
                },
            },
        ],
        [LiteralNode(data=Path("/tmp/foo.txt")), str(Path("/tmp/foo.txt"))],
    ],
)
def test_decode(node: Node, expected: Any):
    assert not DeepDiff(decode(node), expected)



from pathlib import Path
from pipelime.sequences import Sample
import typing as t


class TestAugmentationStages:
    def _load_sample(self, idx: int, folder: Path) -> t.Tuple[Sample, Sample, dict]:
        from pipelime.items import PngImageItem

        items = {}
        items_aug = {}
        keys2trg = {}
        for p in folder.glob(f"{idx}_*_image.png"):
            item_key = p.stem.split("_")[1]
            items[item_key] = PngImageItem(p)
            items_aug[item_key] = PngImageItem(p.parent / f"{p.stem}_aug.png")
            keys2trg[item_key] = "image"
        for p in folder.glob(f"{idx}_*_mask.png"):
            item_key = p.stem.split("_")[1]
            items[item_key] = PngImageItem(p)
            items_aug[item_key] = PngImageItem(p.parent / f"{p.stem}_aug.png")
            keys2trg[item_key] = "mask"
        return Sample(items), Sample(items_aug), keys2trg

    def _stage_albumentations_test_helper(self, folder, transform, outkey, outkey_fm):
        from pipelime.stages import StageAlbumentations
        import numpy as np

        sample, sample_gt, keys2trg = self._load_sample(0, folder)
        stage = StageAlbumentations(
            transform=transform, keys_to_targets=keys2trg, output_key_format=outkey
        )

        sample_aug = stage(sample)
        for key in keys2trg:
            keyaug = outkey_fm.replace("*", key)
            assert key in sample
            assert key in sample_gt
            assert key in sample_aug
            assert keyaug not in sample
            assert keyaug not in sample_gt
            assert keyaug in sample_aug

            assert np.array_equal(sample[key](), sample_aug[key](), equal_nan=True)
            assert np.array_equal(
                sample_gt[key](), sample_aug[keyaug](), equal_nan=True
            )

    def _stage_albumentations_test(self, folder, transform):
        self._stage_albumentations_test_helper(folder, transform, "img-*_", "img-*_")
        self._stage_albumentations_test_helper(folder, transform, "Aug", "*Aug")

    def test_albumentation_object(self, augmentations_folder: Path):
        import albumentations as A

        tr = A.Compose(
            [
                A.Resize(
                    height=50, width=50, interpolation=1, always_apply=False, p=1.0
                ),
                A.Equalize(mode="cv", by_channels=True, always_apply=False, p=1.0),
            ]
        )
        self._stage_albumentations_test(augmentations_folder, tr)

    def test_albumentation_json(self, augmentations_folder: Path):
        self._stage_albumentations_test(
            augmentations_folder, augmentations_folder / "albumentations.json"
        )

    def test_albumentation_dict(self, augmentations_folder: Path):
        dict_tr = {
            "__version__": "1.0.3",
            "transform": {
                "__class_fullname__": "Compose",
                "p": 1.0,
                "transforms": [
                    {
                        "__class_fullname__": "Resize",
                        "always_apply": False,
                        "p": 1,
                        "height": 50,
                        "width": 50,
                        "interpolation": 1,
                    },
                    {
                        "__class_fullname__": "Equalize",
                        "always_apply": False,
                        "p": 1,
                        "mode": "cv",
                        "by_channels": True,
                    },
                ],
                "bbox_params": None,
                "keypoint_params": None,
                "additional_targets": {},
            },
        }
        self._stage_albumentations_test(augmentations_folder, dict_tr)



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "pipelime"
copyright = "2022, Eyecan.ai https://www.eyecan.ai/"
author = "Eyecan.ai"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "myst_parser",
    "sphinxcontrib.mermaid",
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
]

# templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]
autosummary_generate = True
myst_heading_anchors = 3

# -- Options for HTML output -------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_material"
html_logo = "pipelime_logo.svg"
# html_static_path = ["_static"]
html_sidebars = {
    "**": ["logo-text.html", "globaltoc.html", "localtoc.html", "searchbox.html"]
}

# Theme options are theme-specific and customize the look and feel of a
# theme further.  For a list of options available for each theme, see the
# documentation.
# For the icon search for material icons on Google and look in this list
# https://gist.github.com/albionselimaj/14fabdb89d7893c116ee4b48fdfdc7ae
# if there's a valid code for your choice
html_theme_options = {
    "base_url": "https://www.eyecan.ai",
    "repo_url": "https://github.com/eyecan-ai/pypelime-python/",
    "repo_name": "Check pipelime on github!",
    "repo_type": "github",
    "html_minify": True,
    "touch_icon": "pipelime_logo.svg",
    "css_minify": True,
    "nav_title": "pipelime",
    "nav_links": [
        {
            "href": "https://www.eyecan.ai",
            "internal": False,
            "title": "eyecan.ai",
        },
    ],
    "globaltoc_depth": 1,
    "color_primary": "purple",
    "color_accent": "lime",
    # Visible levels of the global TOC; -1 means unlimited
    "globaltoc_depth": 1,
    # If False, expand all TOC entries
    "globaltoc_collapse": False,
    # If True, show hidden TOC entries
    "globaltoc_includehidden": False,
}



import pytest
from pathlib import Path
from urllib.parse import ParseResult
import typing as t

from pipelime.sequences import Sample, SamplesSequence
from pipelime.remotes import make_remote_url


class TestRemotes:
    def _normalized_url(self, url: ParseResult) -> str:
        src_path = Path(url.path)
        if src_path.suffix:
            src_path = src_path.parent
        return ParseResult(
            scheme=url.scheme,
            netloc=url.netloc if url.netloc else "localhost",
            path=src_path.as_posix(),
            params="",
            query="",
            fragment="",
        ).geturl()

    def _upload_to_remote(  # noqa: C901
        self,
        inpath: Path,
        outpath: Path,
        remote_urls: t.Union[ParseResult, t.Sequence[ParseResult]],
        filter_fn: t.Optional[t.Callable[[Sample], bool]],
        keys_to_upload: t.Optional[t.Sequence[str]],
        check_data: bool,
    ):
        from pipelime.stages import StageUploadToRemote
        import pipelime.items as pli
        import numpy as np

        if isinstance(remote_urls, ParseResult):
            remote_urls = [remote_urls]

        filtered_seq = SamplesSequence.from_underfolder(  # type: ignore
            inpath, merge_root_items=False
        )
        if filter_fn:
            filtered_seq = filtered_seq.filter(filter_fn)

        out_seq = filtered_seq.map(
            StageUploadToRemote(remotes=remote_urls, keys_to_upload=keys_to_upload)
        ).to_underfolder(folder=outpath)
        for _ in out_seq:
            pass

        # remove user:password from netloc
        remote_urls = [
            ParseResult(
                scheme=r.scheme,
                netloc=r.netloc.rpartition("@")[2],
                path=r.path,
                params=r.params,
                query=r.query,
                fragment=r.fragment,
            )
            for r in remote_urls
        ]

        # high-level check
        if check_data:
            org_seq = SamplesSequence.from_underfolder(  # type: ignore
                inpath, merge_root_items=False
            )
            if filter_fn:
                org_seq = org_seq.filter(filter_fn)
            reader_out = SamplesSequence.from_underfolder(  # type: ignore
                outpath, merge_root_items=False
            )
            for org_sample, out_sample in zip(org_seq, reader_out):
                assert org_sample.keys() == out_sample.keys()

                for key, out_item in out_sample.items():
                    assert out_item._data_cache is None
                    if not keys_to_upload or key in keys_to_upload:
                        actual_remotes = set(
                            org_sample[key]._remote_sources + remote_urls
                        )

                        assert len(out_item._file_sources) == 0
                        assert len(out_item._remote_sources) == len(actual_remotes)
                        print(out_item._remote_sources)
                        print(actual_remotes)
                        for rm_src in out_item._remote_sources:
                            for rm_trg in actual_remotes:
                                if self._normalized_url(rm_src) == self._normalized_url(
                                    rm_trg
                                ):
                                    break
                            else:
                                assert False
                    else:
                        assert len(out_item._file_sources) == 1
                        assert len(out_item._remote_sources) == 0

                for k, v in org_sample.items():
                    if isinstance(v, pli.NumpyItem):
                        assert np.array_equal(
                            v(), out_sample[k](), equal_nan=True  # type: ignore
                        )
                    else:
                        assert v() == out_sample[k]()

    def test_shared_folder_upload(
        self, minimnist_private_dataset: dict, tmp_path: Path
    ):
        # data lake
        remote_url = make_remote_url(
            scheme="file",
            host="localhost",
            bucket=(tmp_path / "rmbucket"),
        )

        self._upload_to_remote(
            minimnist_private_dataset["path"],
            tmp_path / "output",
            remote_url,
            None,
            None,
            True,
        )

    def test_s3_upload(
        self, minimnist_private_dataset: dict, tmp_path: Path, minio: str
    ):
        # data lake
        if not minio:
            pytest.skip("MinIO unavailable")

        remote_url = make_remote_url(
            scheme="s3",
            user=f"{minio}",
            password=f"{minio}",
            host="localhost",
            port=9000,
            bucket="rmbucket",
            secure=False,
        )

        self._upload_to_remote(
            minimnist_private_dataset["path"],
            tmp_path / "output",
            remote_url,
            None,
            None,
            True,
        )

    def test_incremental_file_upload(
        self, minimnist_private_dataset: dict, tmp_path: Path
    ):
        import pipelime.items as pli
        from shutil import rmtree
        import numpy as np

        # data lake
        remote_root = tmp_path / "file_remote"
        remote_root.mkdir()
        remote_url = make_remote_url(
            scheme="file", host="localhost", bucket=remote_root / "rmbucket"
        )

        # upload even samples
        even_output = tmp_path / "even_uploaded"
        self._upload_to_remote(
            minimnist_private_dataset["path"],
            even_output,
            remote_url,
            lambda x: int(x["label"]()) % 2 == 0,  # type: ignore
            minimnist_private_dataset["image_keys"],
            True,
        )

        # manually merge odd and even samples
        even_odd_partial_output = tmp_path / "even_odd_partial"

        even_seq = SamplesSequence.from_underfolder(  # type: ignore
            even_output, merge_root_items=False
        )
        even_count = len(even_seq)
        assert even_count == minimnist_private_dataset["len"] // 2

        even_odd_partial_seq = even_seq.cat(
            SamplesSequence.from_underfolder(  # type: ignore
                minimnist_private_dataset["path"], merge_root_items=False
            ).filter(lambda x: int(x["label"]()) % 2 == 1)
        )

        # NB: `to_underfolder` generates a new sequence, but Items are shallow-copied.
        # When serializing to disk w/o uploading to remote, the original items are
        # updated with the new sources, since the operation is inherently safe.
        # Therefore, items in `even_odd_partial_seq` will have 2 file sources.
        for _ in even_odd_partial_seq.to_underfolder(folder=even_odd_partial_output):
            pass

        # clear all remote data
        rmtree(remote_root, ignore_errors=True)
        remote_root.mkdir()

        # upload all samples, but only the odd ones are actually copied to the remote
        even_odd_output = tmp_path / "even_odd_uploaded"
        self._upload_to_remote(
            even_odd_partial_output,
            even_odd_output,
            remote_url,
            None,
            minimnist_private_dataset["image_keys"],
            False,
        )

        # only the odd samples are on the remote
        even_odd_reader = SamplesSequence.from_underfolder(  # type: ignore
            even_odd_output, merge_root_items=False
        )
        assert len(even_odd_reader) == minimnist_private_dataset["len"]

        for s1, s2 in zip(even_odd_reader, even_odd_partial_seq):
            is_even = int(s1["label"]()) % 2 == 0
            assert s1.keys() == s2.keys()

            for k, v1 in s1.items():
                v2 = s2[k]
                assert v1.__class__ == v2.__class__

                if (k in minimnist_private_dataset["image_keys"]) and is_even:
                    assert v1() is None
                    assert v2() is None
                    assert len(v1._file_sources) == 0
                    assert len(v1._remote_sources) == 1
                    assert len(v1._file_sources) == len(v2._file_sources)
                    assert len(v1._remote_sources) == len(v2._remote_sources)
                else:
                    assert v2() is not None
                    if k in minimnist_private_dataset["image_keys"]:
                        assert len(v1._file_sources) == 0
                        assert len(v1._remote_sources) == 1
                    else:
                        assert len(v1._file_sources) == 1
                        assert len(v1._remote_sources) == 0
                    # NB: when we called `to_underfolder` a new file source has been
                    # added (see above)
                    assert len(v2._file_sources) == 2
                    assert len(v2._remote_sources) == 0
                    if isinstance(v1, pli.NumpyItem):
                        assert np.array_equal(
                            v1(), v1(), equal_nan=True  # type: ignore
                        )
                    else:
                        assert v1() == v2()

    def test_multiple_remote_upload(
        self, minimnist_private_dataset: dict, tmp_path: Path
    ):
        from shutil import rmtree
        import numpy as np
        import pipelime.items as pli

        # create two remotes
        remote_a_root = tmp_path / "remote_a"
        remote_a_root.mkdir()
        remote_a_url = make_remote_url(
            scheme="file", host="localhost", bucket=remote_a_root / "rmbucketa"
        )

        remote_b_root = tmp_path / "remote_b"
        remote_b_root.mkdir()
        remote_b_url = make_remote_url(
            scheme="file", host="localhost", bucket=remote_b_root / "rmbucketb"
        )

        # upload to both remotes
        output_a_and_b = tmp_path / "output_a_and_b"
        self._upload_to_remote(
            minimnist_private_dataset["path"],
            output_a_and_b,
            [remote_a_url, remote_b_url],
            None,
            None,
            True,
        )

        # now remove remote_a and check again the data
        rmtree(remote_a_root, ignore_errors=True)

        input_seq = SamplesSequence.from_underfolder(  # type: ignore
            minimnist_private_dataset["path"], merge_root_items=False
        )
        output_seq = SamplesSequence.from_underfolder(output_a_and_b)  # type: ignore
        for ins, outs in zip(input_seq, output_seq):
            assert ins.keys() == outs.keys()
            for k, v in ins.items():
                if isinstance(v, pli.NumpyItem):
                    assert np.array_equal(
                        v(), outs[k](), equal_nan=True  # type: ignore
                    )
                else:
                    assert v() == outs[k]()

        # now upload to remote_c taking data from remote_b
        remote_c_root = tmp_path / "remote_c"
        remote_c_root.mkdir()
        remote_c_url = make_remote_url(
            scheme="file", host="localhost", bucket=remote_c_root / "rmbucketc"
        )

        output_c_from_b = tmp_path / "output_c_from_b"
        self._upload_to_remote(
            output_a_and_b,
            output_c_from_b,
            remote_c_url,
            None,
            None,
            True,
        )

    def test_forget_source(self, minimnist_private_dataset: dict, tmp_path: Path):
        from pipelime.stages import StageForgetSource
        import pipelime.items as pli
        import numpy as np

        # create two remotes
        remote_a_root = tmp_path / "remote_a"
        remote_a_root.mkdir()
        remote_a_url = make_remote_url(
            scheme="file", host="localhost", bucket=remote_a_root / "rmbucketa"
        )

        remote_b_root = tmp_path / "remote_b"
        remote_b_root.mkdir()
        remote_b_url = make_remote_url(
            scheme="file", host="localhost", bucket=remote_b_root / "rmbucketb"
        )

        # upload to both remotes
        key_noup_rm, key_noup_norm, key_up_rm = minimnist_private_dataset["item_keys"][
            0:3
        ]
        output_a_and_b = tmp_path / "output_a_and_b"
        self._upload_to_remote(
            minimnist_private_dataset["path"],
            output_a_and_b,
            [remote_a_url, remote_b_url],
            None,
            [
                k
                for k in minimnist_private_dataset["item_keys"]
                if k != key_noup_rm and k != key_noup_norm
            ],
            True,
        )

        # remove remote_a from all samples and remote_b from only one item key
        output_remove = tmp_path / "output_remove"
        seq = (
            SamplesSequence.from_underfolder(output_a_and_b)  # type: ignore
            .map(
                StageForgetSource(
                    remote_a_url,
                    **{
                        key_noup_rm: [output_a_and_b, remote_b_url],
                        key_up_rm: remote_b_url,
                    },
                )
            )
            .to_underfolder(output_remove)
        )
        for _ in seq:
            pass

        # check final output
        normalized_a = self._normalized_url(remote_a_url)
        normalized_b = self._normalized_url(remote_b_url)
        org_seq = SamplesSequence.from_underfolder(output_a_and_b)  # type: ignore
        out_seq = SamplesSequence.from_underfolder(output_remove)  # type: ignore
        for org_sample, out_sample in zip(org_seq, out_seq):
            assert org_sample.keys() == out_sample.keys()
            for k, vout in out_sample.items():
                if k in (key_noup_rm, key_up_rm):
                    # this item should be a deep copy
                    assert len(vout._remote_sources) == 0
                    path = Path(vout._file_sources[0])
                    assert not path.is_symlink()
                    assert path.is_file()
                    assert path.stat().st_nlink == 1
                elif k == key_noup_norm:
                    # this item should be a hard link
                    assert len(vout._remote_sources) == 0
                    path = Path(vout._file_sources[0])
                    assert not path.is_symlink()
                    assert path.is_file()
                    assert path.stat().st_nlink == 3
                else:
                    norm_rm = [self._normalized_url(u) for u in vout._remote_sources]
                    assert normalized_a not in norm_rm
                    assert normalized_b in norm_rm
                    assert len(vout._file_sources) == 0

                if isinstance(vout, pli.NumpyItem):
                    assert np.array_equal(
                        vout(), org_sample[k](), equal_nan=True  # type: ignore
                    )
                else:
                    assert vout() == org_sample[k]()



from typing import Optional

from pipelime.piper.executors.base import NodesGraphExecutor, WatcherNodesGraphExecutor
from pipelime.piper.executors.naive import NaiveNodesGraphExecutor


class NodesGraphExecutorFactory:
    CLASS_MAP = {
        "NAIVE": NaiveNodesGraphExecutor,
    }

    DEFAULT_EXECUTOR_TYPE = "NAIVE"

    @classmethod
    def get_executor(
        cls, type_: Optional[str] = None, watch: bool = False, **kwargs
    ) -> NodesGraphExecutor:
        executor = cls.CLASS_MAP[(type_ or cls.DEFAULT_EXECUTOR_TYPE).upper()](**kwargs)

        if watch:
            executor = WatcherNodesGraphExecutor(executor)

        return executor



from typing import ContextManager


class ContextManagerList:
    def __init__(self, *args: ContextManager):
        self._cms = args

    def __enter__(self):
        for cm in self._cms:
            cm.__enter__()

    def __exit__(self, exc_type, exc_value, traceback):
        for cm in self._cms:
            cm.__exit__(exc_type, exc_value, traceback)



from loguru import logger

from pipelime.piper.executors.base import NodesGraphExecutor
from pipelime.piper.graph import DAGNodesGraph


class NaiveNodesGraphExecutor(NodesGraphExecutor):
    def exec(self, graph: DAGNodesGraph, token: str = "") -> bool:
        """Executes the given graph.

        Args:
            graph (DAGNodesGraph): target graph
            token (str, optional): execution token shared among nodes. Defaults to "".

        Raises:
            RuntimeError: if some node fails to execute

        Returns:
            bool: True if the execution succeeds
        """

        for layer in graph.build_execution_stack():
            for node in layer:
                node.command.set_piper_info(token=token, node=node.name)

                logger.debug(f"Executing command: {node.command._piper.node}")

                node.command()
        return True



from copy import deepcopy
from pathlib import Path

from deepdiff import DeepDiff
from schema import Or, Schema, Use

from pipelime.choixe.ast.parser import parse
from pipelime.choixe.utils.io import load
from pipelime.choixe.visitors import process, walk
from pipelime.choixe.visitors.inspector import Inspection
from pipelime.choixe.xconfig import XConfig


class TestXConfig:
    def _copy_test(self, cfg: XConfig):
        cfg_copy = cfg.copy()
        assert not DeepDiff(cfg_copy.decode(), cfg.decode())
        assert cfg_copy.get_schema() == cfg.get_schema()
        assert cfg_copy.get_cwd() == cfg.get_cwd()

    def test_from_file(self, choixe_plain_cfg: Path):
        cfg = XConfig.from_file(choixe_plain_cfg)
        assert not DeepDiff(cfg.decode(), load(choixe_plain_cfg))
        assert cfg.get_cwd() == choixe_plain_cfg.parent
        self._copy_test(cfg)

    def test_from_dict(self, choixe_plain_cfg: Path):
        data = load(choixe_plain_cfg)
        cfg = XConfig(data=data)
        assert not DeepDiff(cfg.decode(), data)
        assert cfg.get_cwd() is None
        self._copy_test(cfg)

    def test_from_nothing(self):
        cfg = XConfig()
        assert not DeepDiff(cfg.decode(), {})
        assert cfg.get_cwd() is None
        self._copy_test(cfg)

    def test_file_io(self, tmp_path: Path, choixe_plain_cfg: Path):
        cfg = XConfig.from_file(choixe_plain_cfg)
        save_path = tmp_path / "config.yml"
        cfg.save_to(save_path)
        re_cfg = XConfig.from_file(save_path)

        assert re_cfg.get_cwd() == save_path.parent
        assert not DeepDiff(re_cfg.decode(), cfg.decode())

    def test_with_schema(self, choixe_plain_cfg: Path):
        cfg = XConfig.from_file(choixe_plain_cfg)
        assert cfg.is_valid()  # No schema: always valid

        schema = Schema(
            {"alice": int, "bob": int, "charlie": [Or(str, {str: Use(int)})]}
        )
        cfg.set_schema(schema)
        assert cfg.get_schema() == schema
        assert cfg.is_valid()
        cfg.validate()
        expected = schema.validate(load(choixe_plain_cfg))
        assert not DeepDiff(cfg.decode(), expected)

    def test_deep_keys(self, choixe_plain_cfg: Path):
        cfg = XConfig.from_file(choixe_plain_cfg)
        assert cfg.deep_get("charlie[2].alpha") == 10.0
        assert cfg.deep_get(["charlie", 2, "beta"]) == 20.0
        assert cfg.deep_get("bob.alpha", default="hello") == "hello"

        cfg.deep_set("charlie[2].alpha", 40, only_valid_keys=False)
        assert cfg.deep_get("charlie[2].alpha") == 40

        cfg.deep_set("charlie[2].alpha", 50, only_valid_keys=True)
        assert cfg.deep_get("charlie[2].alpha") == 50

        cfg.deep_set("charlie[3].foo.bar", [10, 20, 30], only_valid_keys=False)
        assert cfg.deep_get("charlie[3]") == {"foo": {"bar": [10, 20, 30]}}

        cfg.deep_set("charlie[4].foo.bar", [10, 20, 30], only_valid_keys=True)
        assert cfg.deep_get("charlie[4]") is None

        cfg.deep_set("dino[2][3].saur", 42, only_valid_keys=False)
        assert cfg.deep_get("dino[0]") is None
        assert cfg.deep_get("dino[2][0]") is None
        assert cfg.deep_get("dino[2][3].saur") == 42

    def test_deep_update(self):
        data = {
            "a": {"b": 10},
            "b": [0, 2, 1.02],
            "c": {"a": "b", "b": [{"a": 1, "b": 2}, "a"]},
        }
        other = {"c": {"b": [{"a": 2}], "e": {"a": 18, "b": "a"}}}
        cfg = XConfig(data=data)
        cfg.deep_update(other)
        expected = deepcopy(data)
        expected["c"]["b"][0]["a"] = 2
        assert not DeepDiff(cfg.decode(), expected)

    def test_full_merge(self):
        data = {
            "a": {"b": 10},
            "b": [0, 2, 1.02],
            "c": {"a": "b", "b": [{"a": 1, "b": 2}, "a"]},
        }
        other = {"c": {"b": [{"a": 2}], "e": {"a": 18, "b": "a"}}}
        cfg = XConfig(data=data)
        cfg.deep_update(other, full_merge=True)
        expected = {
            "a": {"b": 10},
            "b": [0, 2, 1.02],
            "c": {"a": "b", "b": [{"a": 2, "b": 2}, "a"], "e": {"a": 18, "b": "a"}},
        }
        assert not DeepDiff(cfg.decode(), expected)

    def test_walk(self, choixe_plain_cfg: Path):
        cfg = XConfig.from_file(choixe_plain_cfg)
        assert not DeepDiff(cfg.walk(), walk(parse(load(choixe_plain_cfg))))

    def test_process(self, choixe_plain_cfg: Path):
        cfg = XConfig.from_file(choixe_plain_cfg)
        processed = cfg.process().decode()
        processed_expected = process(
            parse(load(choixe_plain_cfg)), allow_branching=False
        )[0]
        assert not DeepDiff(processed, processed_expected)

    def test_process_all(self, choixe_plain_cfg: Path):
        cfg = XConfig.from_file(choixe_plain_cfg)
        processed = cfg.process_all()
        processed_expected = process(
            parse(load(choixe_plain_cfg)), allow_branching=True
        )
        for a, b in zip(processed, processed_expected):
            assert not DeepDiff(a.decode(), b)

    def test_inspect(self, choixe_plain_cfg: Path):
        cfg = XConfig.from_file(choixe_plain_cfg)
        inspection = cfg.inspect()
        expected = Inspection(processed=True)
        assert inspection == expected



import ast
import re
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, OrderedDict, Tuple, Type, Union

import astunparse
import schema as S

import pipelime.choixe.ast.nodes as c_ast

DIRECTIVE_PREFIX = "$"
"""Prefix used at the start of all Choixe directives."""


class ChoixeParsingError(Exception):
    pass


class ChoixeSyntaxError(ChoixeParsingError):
    pass


class ChoixeTokenValidationError(ChoixeParsingError):
    pass


class ChoixeStructValidationError(ChoixeParsingError):
    pass


@dataclass
class Token:
    name: str
    args: List[Any]
    kwargs: Dict[str, Any]


class Scanner:
    """Choixe Scanner of python str objects."""

    DIRECTIVE_RE = r"(?:\$[^\)\( \.\,\$]+\([^\(\)]*\))|(?:\$[^\)\( \.\,\$]+)|(?:[^\$]*)"
    """Regex used to check if a string is a Choixe directive."""

    def _scan_argument(
        self, py_arg: Union[ast.Constant, ast.Attribute, ast.Name]
    ) -> Any:
        if isinstance(py_arg, ast.Constant):
            return py_arg.value
        elif isinstance(py_arg, ast.Attribute):
            name = astunparse.unparse(py_arg)
            if name.endswith("\n"):  # Remove trailing newline
                name = name[:-1]
            return str(name)
        elif isinstance(py_arg, ast.Name):
            return str(py_arg.id)
        else:
            raise ChoixeSyntaxError(py_arg.__class__)

    def _scan_directive(self, code: str) -> Token:
        try:
            py_ast = ast.parse(f"_{code}")  # Add "_" to avoid conflicts with python
        except SyntaxError:
            raise ChoixeSyntaxError(code)

        assert isinstance(py_ast, ast.Module)
        py_expr = py_ast.body[0].value  # type: ignore

        if isinstance(py_expr, ast.Call):
            # Remove the added "_"
            token_name = py_expr.func.id[1:]  # type: ignore
            py_args = py_expr.args
            py_kwargs = py_expr.keywords
        elif isinstance(py_expr, ast.Name):
            token_name = py_expr.id[1:]
            py_args = []
            py_kwargs = {}
        else:
            raise ChoixeSyntaxError(code)

        args = []
        for py_arg in py_args:
            args.append(self._scan_argument(py_arg))  # type: ignore

        kwargs = {}
        for py_kwarg in py_kwargs:
            key, value = py_kwarg.arg, py_kwarg.value
            kwargs[key] = self._scan_argument(value)  # type: ignore

        return Token(token_name, args, kwargs)

    def scan(self, data: str) -> List[Token]:
        """Transforms a string into a list of parsed tokens.

        Args:
            data (str): The string to parse.

        Returns:
            List[Token]: The list of parsed tokens.
        """
        res = []
        tokens = re.findall(self.DIRECTIVE_RE, data)
        for token in tokens:
            if len(token) == 0:
                continue
            if token.startswith(DIRECTIVE_PREFIX):
                token = self._scan_directive(token[len(DIRECTIVE_PREFIX) :])
            else:
                token = Token("str", [token], {})
            res.append(token)

        return res


class Parser:
    """Choixe parser for all kind of python objects."""

    def __init__(self) -> None:
        self._scanner = Scanner()
        self._type_map = {
            dict: self._parse_dict,
            list: self._parse_list,
            tuple: self._parse_list,
            str: self._parse_str,
        }

        self._call_forms = {
            self._token_schema("var"): c_ast.VarNode,
            self._token_schema("import"): c_ast.ImportNode,
            self._token_schema("sweep"): c_ast.SweepNode,
            self._token_schema("index"): c_ast.IndexNode,
            self._token_schema("item"): c_ast.ItemNode,
            self._token_schema("uuid"): c_ast.UuidNode,
            self._token_schema("date"): c_ast.DateNode,
            self._token_schema("cmd"): c_ast.CmdNode,
            self._token_schema("tmp"): c_ast.TmpDirNode,
            self._token_schema("symbol"): c_ast.SymbolNode,
        }

        self._extended_and_special_forms = {
            S.Schema(
                {
                    self._token_schema("directive"): str,
                    self._token_schema("args"): list,
                    self._token_schema("kwargs"): dict,
                }
            ): self._parse_extended_form,
            S.Schema(
                {
                    self._token_schema("call"): str,
                    S.Optional(self._token_schema("args")): S.Optional(dict),
                }
            ): self._parse_instance,
            S.Schema(
                {
                    self._token_schema("model"): str,
                    S.Optional(self._token_schema("args")): dict,
                }
            ): self._parse_model,
        }

        self._key_value_forms = {
            S.Schema({self._token_schema("for"): object}): self._parse_for,
            S.Schema({self._token_schema("switch"): list}): self._parse_switch,
        }

    def _token_schema(
        self,
        name: str,
        args: Optional[OrderedDict[str, Type]] = None,
        kwargs: Optional[OrderedDict[str, Type]] = None,
    ) -> S.Schema:
        args = OrderedDict() if args is None else args
        kwargs = OrderedDict() if kwargs is None else kwargs

        def _validator(token: Union[str, Token]) -> bool:
            if isinstance(token, str):
                tokens = self._scanner.scan(token)
                if len(tokens) != 1:
                    return False
                token = tokens[0]

            if token.name != name:
                return False

            return True

        return S.Schema(_validator)

    def _key_value_pairs_by_token_name(
        self, data: Dict[str, Any]
    ) -> Dict[str, Tuple[Token, Any]]:
        res = {}
        for k, v in data.items():
            token = self._scanner.scan(k)[0]
            res[token.name] = (token, v)
        return res

    def _parse_extended_form(self, data: dict) -> c_ast.Node:
        pairs = self._key_value_pairs_by_token_name(data)
        token = Token(pairs["directive"][1], pairs["args"][1], pairs["kwargs"][1])
        return self._parse_token(token)

    def _parse_instance(self, data: dict) -> c_ast.InstanceNode:
        pairs = self._key_value_pairs_by_token_name(data)
        symbol = self.parse(pairs["call"][1])
        args = (
            self.parse(pairs["args"][1])
            if "args" in pairs
            else c_ast.DictNode(nodes={})
        )
        return c_ast.InstanceNode(symbol=symbol, args=args)  # type: ignore

    def _parse_model(self, data: dict) -> c_ast.ModelNode:
        pairs = self._key_value_pairs_by_token_name(data)
        symbol = self.parse(pairs["model"][1])
        args = (
            self.parse(pairs["args"][1])
            if "args" in pairs
            else c_ast.DictNode(nodes={})
        )
        return c_ast.ModelNode(symbol, args)  # type: ignore

    def _parse_for(self, loop: str, body: Any) -> c_ast.ForNode:
        token = self._scanner.scan(loop)[0]
        iterable = c_ast.LiteralNode(data=token.args[0])
        identifier = (
            token.args[1] if len(token.args) > 1 else token.kwargs.get("identifier")
        )
        identifier = c_ast.LiteralNode(data=identifier) if identifier else None
        return c_ast.ForNode(
            iterable=iterable, body=self.parse(body), identifier=identifier
        )

    def _parse_switch(self, switch: str, body: Any) -> c_ast.SwitchNode:
        token = self._scanner.scan(switch)[0]
        value = c_ast.LiteralNode(data=token.args[0])
        cases = []
        default = None
        for entry in body:
            pairs = self._key_value_pairs_by_token_name(entry)
            if "default" in pairs:
                default = self.parse(pairs["default"][1])
            elif "case" in pairs:
                case_set = self.parse(pairs["case"][1])
                case_body = self.parse(pairs["then"][1])
                cases.append((case_set, case_body))
        return c_ast.SwitchNode(value=value, cases=cases, default=default)

    def _parse_dict(self, data: dict) -> Union[c_ast.DictNode, c_ast.DictBundleNode]:
        # Check if the dict is an extended or special form, in that case parse it and
        # return the result, no further checks are needed.
        for schema, fn in self._extended_and_special_forms.items():
            if schema.is_valid(data):
                try:
                    return fn(data)
                except Exception as e:
                    raise ChoixeStructValidationError(data) from e

        # Check for each entry if it is a key_value form, in that case parse it and add
        # it to the bundle. Keep track of what is left to parse.
        parsed_keyvalues = []
        parsed_other = {}
        for k, v in data.items():
            any_valid = False

            for schema, fn in self._key_value_forms.items():
                if schema.is_valid({k: v}):
                    parsed_keyvalues.append(fn(k, v))
                    any_valid = True
                    break

            if not any_valid:
                parsed_other[self._parse_str(k)] = self.parse(v)

        # Parse the remaining entries as a DictNode.
        parsed_other = c_ast.DictNode(nodes=parsed_other)

        # If no key_value form was found, return the DictNode.
        if len(parsed_keyvalues) == 0:
            return parsed_other
        # If there is only one key_value form and no remaining data, return it.
        elif len(parsed_other.nodes) == 0 and len(parsed_keyvalues) == 1:
            return parsed_keyvalues[0]
        # If there is more than one key_value form, return a DictBundleNode.
        else:
            return c_ast.DictBundleNode(*parsed_keyvalues, parsed_other)

    def _parse_list(self, data: list) -> c_ast.ListNode:
        return c_ast.ListNode(*[self.parse(x) for x in data])

    def _parse_token(self, token: Token) -> c_ast.Node:
        if token.name == "str":
            return c_ast.LiteralNode(data=token.args[0])

        try:
            for schema, fn in self._call_forms.items():
                if schema.is_valid(token):
                    args = [self.parse(x) for x in token.args]
                    kwargs = {k: self.parse(v) for k, v in token.kwargs.items()}
                    node = fn(*args, **kwargs)
                    return node
        except TypeError:
            raise ChoixeStructValidationError(token)

        raise ChoixeTokenValidationError(token)

    def _parse_str(self, data: str) -> c_ast.Node:
        nodes = []
        for token in self._scanner.scan(data):
            nodes.append(self._parse_token(token))

        if len(nodes) == 0:
            return c_ast.LiteralNode(data="")
        elif len(nodes) == 1:
            return nodes[0]
        else:
            return c_ast.StrBundleNode(*nodes)

    def parse(self, data: Any) -> c_ast.Node:
        """Recursively transforms an object into a visitable AST node.

        Args:
            data (Any): The object to parse.

        Returns:
            Node: The parsed Choixe AST node.
        """
        try:
            fn = c_ast.LiteralNode
            for type_, parse_fn in self._type_map.items():
                if isinstance(data, type_):
                    fn = parse_fn
                    break
            res = fn(data)  # type: ignore
            return res

        except (ChoixeSyntaxError, SyntaxError) as e:
            code = e.args[0]
            raise ChoixeParsingError(
                f'Error when parsing code "{code}" found in "{data}", expected either '
                "a compact form like $DIRECTIVE, or a call form like "
                "$DIRECTIVE(ARGS, KWARGS)."
            )
        except ChoixeTokenValidationError as e:
            token: Token = e.args[0]
            raise ChoixeParsingError(
                f'Token "{token}" found in "{data}" does not validate against any of '
                "the available call forms. Please check that the directive and "
                "argument names are spelled correctly and match the directive "
                "signature."
            )
        except ChoixeStructValidationError as e:
            expr = e.args[0]
            raise ChoixeParsingError(
                f'Structure "{expr}" does not validate against any of the available '
                "special or extended forms. Please check that the directive and "
                "argument names are spelled correctly and match the directive "
                "signature."
            )


def parse(data: Any) -> c_ast.Node:
    """Recursively transforms an object into a visitable AST node.

    Args:
        data (Any): The object to parse.

    Returns:
        Node: The parsed Choixe AST node.
    """
    return Parser().parse(data)



import os
import tempfile
import uuid
from copy import deepcopy
from dataclasses import dataclass
from datetime import datetime
from itertools import product
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional

import pydash as py_

import pipelime.choixe.ast.nodes as ast
from pipelime.choixe.ast.parser import parse
from pipelime.choixe.utils.imports import import_symbol
from pipelime.choixe.utils.io import load
from pipelime.choixe.visitors.unparser import unparse


class ChoixeProcessingError(Exception):
    pass


@dataclass
class LoopInfo:
    index: int
    item: Any


class Processor(ast.NodeVisitor):
    """`NodeVisitor` that implements the processing logic of Choixe."""

    def __init__(
        self,
        context: Optional[Dict[str, Any]] = None,
        cwd: Optional[Path] = None,
        allow_branching: bool = True,
    ) -> None:
        """Constructor for `Processor`

        Args:
            context (Optional[Dict[str, Any]], optional): A data structure containing
            the values that will replace the variable nodes. Defaults to None.
            cwd (Optional[Path], optional): current working directory used for relative
            imports. If set to None, the `os.getcwd()` will be used. Defaults to None.
            allow_branching (bool, optional): Set to False to disable processing on
            branching nodes, like sweeps. All branching nodes will be simply unparsed.
            Defaults to True.
        """
        super().__init__()
        self._context = context if context is not None else {}
        self._cwd = cwd if cwd is not None else Path(os.getcwd())
        self._allow_branching = allow_branching

        self._loop_data: Dict[str, LoopInfo] = {}
        self._current_loop: Optional[str] = None
        self._tmp_name = uuid.uuid1().hex

    def _branches(self, *branches: List[Any]) -> List[Any]:
        if len(branches) == 1:
            return branches[0]
        return list(product(*branches))

    def _repeat(self, data: Any, n: int) -> List[Any]:
        new_data = []
        for _ in range(n):
            new_data.extend(deepcopy(data))
        return new_data

    def visit_dict(self, node: ast.DictNode) -> List[Dict]:
        data = [{}]
        for k, v in node.nodes.items():
            branches = self._branches(k.accept(self), v.accept(self))
            new_data = self._repeat(data, len(branches))
            for i, d in enumerate(new_data):
                d[branches[i // len(data)][0]] = branches[i // len(data)][1]
            data = new_data
        return data

    def visit_list(self, node: ast.ListNode) -> List[List]:
        data = [[]]
        for x in node.nodes:
            branches = x.accept(self)
            new_data = self._repeat(data, len(branches))
            for i, d in enumerate(new_data):
                d.append(branches[i // len(data)])
            data = new_data
        return data

    def visit_literal(self, node: ast.LiteralNode) -> List[Any]:
        return [node.data]

    def visit_dict_bundle(self, node: ast.DictBundleNode) -> List[Dict]:
        data = [{}]
        for x in node.nodes:
            branches = x.accept(self)
            new_data = self._repeat(data, len(branches))
            for i in range(len(new_data)):
                new_data[i].update(branches[i // len(data)])
            data = new_data
        return data

    def visit_str_bundle(self, node: ast.StrBundleNode) -> List[str]:
        data = [""]
        for x in node.nodes:
            branches = x.accept(self)
            N = len(data)
            data *= len(branches)
            for i in range(len(data)):
                data[i] += str(branches[i // N])
        return data

    def visit_var(self, node: ast.VarNode) -> List[Any]:
        id_branches = node.identifier.accept(self)
        default_brances = node.default.accept(self) if node.default else [None]
        env_branches = node.env.accept(self) if node.env else [None]

        branches = self._branches(id_branches, default_brances, env_branches)
        data = []

        for id_, default, env in branches:
            if py_.has(self._context, id_):
                data.append(py_.get(self._context, id_))
                continue

            if env:
                value = os.getenv(id_)
                if value is not None:
                    data.append(value)
                    continue

            if default is not None:
                data.append(default)
                continue

            raise ChoixeProcessingError(f"Variable not found: `{id_}`")

        return data

    def visit_import(self, node: ast.ImportNode) -> List[Any]:
        all_nested = []
        branches = node.path.accept(self)
        for path in branches:
            path = Path(path)
            if not path.is_absolute():
                path = self._cwd / path

            subdata = load(path)
            parsed = parse(subdata)

            old_cwd = self._cwd
            self._cwd = path.parent
            nested = parsed.accept(self)
            self._cwd = old_cwd

            all_nested.append(nested)

        return self._branches(*all_nested)

    def visit_sweep(self, node: ast.SweepNode) -> List[Any]:
        if self._allow_branching:
            cases = []
            for x in node.cases:
                cases.extend(x.accept(self))
            return cases
        else:
            return [unparse(node)]

    def visit_symbol(self, node: ast.SymbolNode) -> List[Any]:
        branches = node.symbol.accept(self)
        return [import_symbol(s, cwd=self._cwd) for s in branches]

    def visit_instance(self, node: ast.InstanceNode) -> List[Any]:
        symbol_branches = node.symbol.accept(self)
        args_branches = node.args.accept(self)
        branches = self._branches(symbol_branches, args_branches)
        return [import_symbol(s, cwd=self._cwd)(**a) for s, a in branches]

    def visit_model(self, node: ast.ModelNode) -> Any:
        symbol_branches = node.symbol.accept(self)
        args_branches = node.args.accept(self)
        branches = self._branches(symbol_branches, args_branches)
        return [import_symbol(s, cwd=self._cwd).parse_obj(a) for s, a in branches]

    def visit_for(self, node: ast.ForNode) -> List[Any]:
        if isinstance(node.iterable.data, int):
            iterable = list(range(node.iterable.data))
        else:
            iterable = py_.get(self._context, node.iterable.data)

        if isinstance(iterable, int):
            iterable = list(range(iterable))

        if not isinstance(iterable, Iterable):
            if not py_.has(self._context, node.iterable.data):
                raise ChoixeProcessingError(
                    f"Loop variable `{node.iterable.data}` not found in context"
                )
            raise ChoixeProcessingError(
                f"Loop variable `{node.iterable.data}` is not iterable"
            )
        id_ = uuid.uuid1().hex if node.identifier is None else str(node.identifier.data)
        prev_loop = self._current_loop
        self._current_loop = id_

        branches: List[Any] = []
        for i, x in enumerate(iterable):
            self._loop_data[self._current_loop] = LoopInfo(i, x)
            branches.append(node.body.accept(self))

        self._current_loop = prev_loop

        branches = self._branches(*branches)
        for i, branch in enumerate(branches):
            if isinstance(node.body, ast.LiteralNode) or isinstance(
                node.body, ast.StrBundleNode
            ):
                res = "".join([str(item) for item in branch])
            elif isinstance(node.body, ast.ListNode):
                res = []
                [res.extend(item) for item in branch]
            else:
                res = {}
                [res.update(item) for item in branch]
            branches[i] = res

        return branches

    def visit_switch(self, node: ast.SwitchNode) -> List[Any]:
        value_branches = node.value.accept(self)
        set_branches = [x[0].accept(self) for x in node.cases]
        body_branches = [x[1].accept(self) for x in node.cases]
        default_branches = node.default.accept(self) if node.default else [None]

        all_branches = self._branches(
            value_branches, *set_branches, *body_branches, default_branches
        )

        branches = []
        for branch in all_branches:
            value = py_.get(self._context, branch[0])
            found = False
            for i in range(len(node.cases)):
                set_ = branch[i + 1]
                if not isinstance(set_, Iterable) or isinstance(set_, str):
                    set_ = [set_]

                if value in set_:
                    branches.append(branch[i + 1 + len(node.cases)])
                    found = True
                    break
            if not found and branch[-1] is not None:
                branches.append(branch[-1])

        return branches

    def visit_index(self, node: ast.IndexNode) -> List[Any]:
        branches = (
            node.identifier.accept(self) if node.identifier else [self._current_loop]
        )
        return [self._loop_data[x].index for x in branches]  # type: ignore

    def visit_item(self, node: ast.ItemNode) -> List[Any]:
        branches = (
            node.identifier.accept(self) if node.identifier else [self._current_loop]
        )
        sep = "."
        items = []
        for branch in branches:
            loop_id, _, key = str(branch).partition(sep)  # type: ignore
            item = py_.get(self._loop_data[loop_id].item, f"{sep}{key}")
            items.append(item)
        return items

    def visit_uuid(self, node: ast.UuidNode) -> List[str]:
        return [uuid.uuid1().hex]

    def visit_date(self, node: ast.DateNode) -> List[str]:
        ts = datetime.now()
        branches = node.format.accept(self) if node.format else [None]
        dates = []
        for branch in branches:
            date = ts.strftime(branch) if branch else ts.isoformat()
            dates.append(date)
        return dates

    def visit_cmd(self, node: ast.CmdNode) -> List[str]:
        stdouts = []
        for branch in node.command.accept(self):
            subp = os.popen(str(branch))
            stdout = subp.read()
            stdouts.append(stdout)
        return stdouts

    def visit_tmp_dir(self, node: ast.TmpDirNode) -> Any:
        paths = []
        branches = node.name.accept(self) if node.name else [self._tmp_name]
        for branch in branches:
            path = Path(tempfile.gettempdir()) / str(branch)
            path.parent.mkdir(exist_ok=True, parents=True)
            paths.append(str(path))
        return paths


def process(
    node: ast.Node,
    context: Optional[Dict[str, Any]] = None,
    cwd: Optional[Path] = None,
    allow_branching: bool = True,
) -> Any:
    """Processes a Choixe AST node into a list of all possible outcomes.

    Args:
        node (Node): The AST node to process.
        context (Optional[Dict[str, Any]], optional): A data structure containing
        the values that will replace the variable nodes. Defaults to None.
        cwd (Optional[Path], optional): current working directory used for relative
        imports. If set to None, the `os.getcwd()` will be used. Defaults to None.
        allow_branching (bool, optional): Set to False to disable processing on
        branching nodes, like sweeps. All branching nodes will be simply unparsed.
        Defaults to True.

    Returns:
        Any: The list of all possible outcomes. If branching is disabled, the list will
        have length 1.
    """
    processor = Processor(context=context, cwd=cwd, allow_branching=allow_branching)
    return node.accept(processor)



from pydantic import Field
import numpy as np

import pipelime.sequences as pls
import pipelime.stages as plst


class AverageColor(plst.SampleStage, title="avg_color"):
    """Average the color of an image."""

    image_key: str = Field("image", description="The key of the image to average.")
    avg_key: str = Field("avg_color", description="The key of the average color.")

    def __call__(self, x: pls.Sample) -> pls.Sample:
        return x.set_value(self.avg_key, np.mean(x[self.image_key](), axis=(0, 1)))  # type: ignore



from itertools import count
from typing import Iterable, Optional, Sequence, Union

from pipelime.piper.progress.model import OperationInfo, ProgressUpdate


class TrackCallback:
    """A custom callback on start, stop and advancement of a running operation"""

    def __init__(self):
        """Constructor for a generic `TrackCallback`"""
        self._op_info = None
        self._progress = 0

    @property
    def _ready(self) -> bool:
        return self._op_info is not None

    def start(self, op_info: OperationInfo) -> None:
        """Set this callback in "ready" state and call the `on_start` callback

        Args:
            op_info (OperationInfo): The operation info to track

        Raises:
            RuntimeError: If the callback is already setup
        """
        if self._ready:
            raise RuntimeError("Callback already setup")
        self._op_info = op_info
        self._progress = 0
        prog = ProgressUpdate(op_info=self._op_info, progress=self._progress)
        self.on_start(prog)

    def advance(self, advance: int = 1) -> None:
        """Advance the progress of the tracked operation by a custom amount of steps

        Args:
            advance (int, optional): The number of steps to advance. Defaults to 1.

        Raises:
            RuntimeError: If the callback is not setup
        """
        if not self._ready:
            raise RuntimeError("Callback not setup")
        self._progress += advance
        prog = ProgressUpdate(op_info=self._op_info, progress=self._progress)
        self.on_advance(prog)

    def finish(self) -> None:
        """Finish the trackeing of the operation and call the `on_finish` callback

        Raises:
            RuntimeError: If the callback is not setup
        """
        if not self._ready:
            raise RuntimeError("Callback not setup")
        op_info = self._op_info
        self._progress = op_info.total
        prog = ProgressUpdate(op_info=op_info, finished=True, progress=self._progress)
        self.on_finish(prog)
        self._op_info = None

    def on_start(self, prog: ProgressUpdate) -> None:
        """What to do when the operation is started

        Args:
            prog (ProgressUpdate): The progress update object.
        """
        pass

    def on_advance(self, prog: ProgressUpdate) -> None:
        """What to do when the operation advances

        Args:
            prog (ProgressUpdate): The progress update object.
        """
        pass

    def on_finish(self, prog: ProgressUpdate) -> None:
        """What to do when the operation is finished

        Args:
            prog (ProgressUpdate): The progress update object.
        """
        pass


class Tracker:
    """Tracker for running operations"""

    def __init__(self, token: str, node: str, *callbacks: TrackCallback) -> None:
        """Constructor for a `Tracker`

        Args:
            token (str): The piper token
            node (str): The piper node name
            callbacks (TrackCallback, optional): The callbacks to use.
        """
        self._counter = count()
        self._callbacks = callbacks
        self._token = token
        self._node = node

    def track(
        self,
        seq: Union[Iterable, Sequence],
        size: Optional[int] = None,
        message: str = "",
    ) -> Iterable:
        """Track a generic iterable sequence"""

        id_ = next(self._counter)
        op_info = OperationInfo(
            token=self._token,
            node=self._node,
            chunk=id_,
            total=len(seq) if size is None else size,  # type: ignore
            message=message,
        )

        for callback in self._callbacks:
            callback.start(op_info)

        for x in seq:
            yield x

            for callback in self._callbacks:
                callback.advance()

        for callback in self._callbacks:
            callback.finish()



from typing import Optional

from pipelime.piper.progress.tracker.base import TrackCallback
from pipelime.piper.progress.tracker.loguru import LoguruTrackCallback
from pipelime.piper.progress.tracker.zmq import ZmqTrackCallback


class TrackCallbackFactory:
    """Factory for ``TrackCallback`` s"""

    DEFAULT_CALLBACK_TYPE = "ZMQ"

    CLASS_MAP = {
        "ZMQ": ZmqTrackCallback,
        "LOGURU": LoguruTrackCallback,
    }

    @classmethod
    def get_callback(cls, type_: Optional[str] = None, **kwargs) -> TrackCallback:
        return cls.CLASS_MAP[(type_ or cls.DEFAULT_CALLBACK_TYPE).upper()](**kwargs)



from typing import Any

import pytest
from pipelime.choixe.ast.nodes import (
    DictNode,
    ListNode,
    Node,
    LiteralNode,
    StrBundleNode,
    VarNode,
)
from pipelime.choixe.visitors import walk
from deepdiff import DeepDiff


@pytest.mark.parametrize(
    ["node", "expected"],
    [
        [LiteralNode(data=10), [([], 10)]],
        [
            StrBundleNode(
                LiteralNode(data="alice "),
                VarNode(
                    identifier=LiteralNode(data="foo"),
                    default=LiteralNode(data="loves"),
                ),
                LiteralNode(data=" bob"),
            ),
            [([], "alice $var(foo, default=loves) bob")],
        ],
        [
            DictNode(
                nodes={
                    LiteralNode(data="key1"): LiteralNode(data=10),
                    LiteralNode(data="key2"): DictNode(
                        nodes={
                            LiteralNode(data="key1"): LiteralNode(data=10.2),
                            LiteralNode(data="key2"): LiteralNode(data="hello"),
                        }
                    ),
                }
            ),
            [(["key1"], 10), (["key2", "key1"], 10.2), (["key2", "key2"], "hello")],
        ],
        [
            ListNode(
                LiteralNode(data=10),
                LiteralNode(data=-0.25),
                ListNode(LiteralNode(data="aa")),
            ),
            [([0], 10), ([1], -0.25), ([2, 0], "aa")],
        ],
    ],
)
def test_walk(node: Node, expected: Any):
    assert not DeepDiff(walk(node), expected)



from abc import ABC, abstractmethod
from threading import Thread
import time
from typing import Optional

from loguru import logger

from pipelime.piper.progress.model import ProgressUpdate


class ListenerCallback:
    """A callback for the listener"""

    def on_start(self) -> None:
        """Called when the listener starts"""
        pass

    def on_update(self, prog: ProgressUpdate) -> None:
        """Called when a progress update is received"""
        pass

    def on_stop(self) -> None:
        """Called when the listener stops"""
        pass


class ProgressReceiver(ABC):
    """A receiver for progress updates"""

    def __init__(self, token: str) -> None:
        super().__init__()
        self._token = token

    @abstractmethod
    def receive(self) -> Optional[ProgressUpdate]:
        """Receive a progress update"""
        pass

    def __next__(self) -> Optional[ProgressUpdate]:
        """Wait for the next progress update"""
        try:
            res = self.receive()
        except Exception as e:  # pragma: no cover
            logger.exception(e)
            res = None
        return res


class Listener:
    """A listener for progress updates"""

    def __init__(
        self,
        receiver: ProgressReceiver,
        *callbacks: ListenerCallback,
        min_poll_interval: float = 0.1
    ) -> None:
        """Initialize the listener

        Args:
            receiver (ProgressReceiver): The progress receiver to use
            callbacks (ListenerCallback): The callbacks to call
            min_poll_interval (float): The minimum interval, in seconds, between
                progress updates
        """
        self._receiver = receiver
        self._callbacks = callbacks
        self._min_poll_interval = min_poll_interval

        self._stop_flag = False
        self._listening_thread = None

    def _listen(self) -> None:
        while not self._stop_flag:
            last_update = time.time()
            prog = next(self._receiver)
            if prog is not None:
                for cb in self._callbacks:
                    cb.on_update(prog)

            time_diff = time.time() - last_update
            if time_diff < self._min_poll_interval:
                time.sleep(time_diff)

    def start(self) -> None:
        """Start the listener in a thread"""
        self._listening_thread = Thread(target=self._listen)
        self._listening_thread.start()

        for cb in self._callbacks:
            cb.on_start()

    def stop(self) -> None:
        """Stop the listener"""
        if self._listening_thread is not None:
            self._stop_flag = True
            self._listening_thread.join(5.0)
            self._listening_thread = None

            for cb in self._callbacks:
                cb.on_stop()



import typing as t
import pydantic as pyd

from pipelime.stages import SampleStage
from pipelime.utils.pydantic_types import ItemType

if t.TYPE_CHECKING:
    from pipelime.sequences import Sample


class StageReplaceItem(SampleStage, title="replace-item"):
    """Replaces items in sample preserving internal values."""

    key_item_map: t.Mapping[str, ItemType] = pyd.Field(
        ...,
        description=(
            "A mapping `key: item_cls` returning the new item class for the key."
        ),
    )

    def __call__(self, x: "Sample") -> "Sample":
        for key, item_cls in self.key_item_map.items():
            if key in x:
                old_item = x[key]
                x = x.set_item(
                    key, item_cls.itype.make_new(old_item, shared=old_item.is_shared)
                )
        return x



from typing import Optional

import zmq

from pipelime.piper.progress.listener.base import ProgressReceiver
from pipelime.piper.progress.model import ProgressUpdate


class ZMQProgressReceiver(ProgressReceiver):
    """A receiver for progress updates over pubsub ZMQ socket"""

    def __init__(self, token: str) -> None:
        self._token = token
        super().__init__(token)
        self._context = zmq.Context()
        self._socket = self._context.socket(zmq.SUB)
        self._socket.connect("tcp://localhost:5556")
        self._socket.subscribe(token.encode())

    def receive(self) -> Optional[ProgressUpdate]:
        if self._socket is not None:
            while True:
                # Receive a message
                try:
                    token, messagedata = self._socket.recv_multipart(flags=zmq.NOBLOCK)
                except zmq.ZMQError:
                    return None  # no message

                # If token is wrong, wait for another message
                if token.decode() == self._token:
                    break

            # Parse message and return
            return ProgressUpdate.parse_raw(messagedata.decode())



from pathlib import Path
from typing import Dict, Optional

from pipelime.choixe import XConfig
from pipelime.piper.model import DAGModel
from pipelime.piper.parsers.base import DAGParser


class ChoixeDAGParser(DAGParser):
    def parse_cfg(self, cfg: Dict, params: Optional[Dict] = None) -> DAGModel:
        cfg = XConfig(cfg)
        context = XConfig(params)

        cfg = cfg.process(context)
        assert cfg.inspect().processed, "The configuration is still not processed"

        return DAGModel.parse_obj(cfg)

    def _read_file(self, path: Path, additional_args: Optional[Dict] = None) -> Dict:
        data = XConfig.from_file(path)
        additional_args = {} if additional_args is None else additional_args
        data.deep_update(additional_args)
        return data



import typing as t

import pydantic as pyd

import pipelime.commands.interfaces as pl_interfaces
from pipelime.piper import PipelimeCommand, PiperPortType


class SplitBase(
    pl_interfaces.PydanticFieldNoDefaultMixin,
    pyd.BaseModel,
    allow_population_by_field_name=True,
    extra="forbid",
    copy_on_model_validation="none",
):
    output: t.Optional[
        pl_interfaces.OutputDatasetInterface
    ] = pl_interfaces.OutputDatasetInterface.pyd_field(
        alias="o",
        is_required=False,
        description="Output split. Leave to None to not save to disk.",
    )

    def __repr__(self) -> str:
        return self.__piper_repr__()

    def __piper_repr__(self) -> str:
        return "" if self.output is None else self.output.__piper_repr__()


class PercSplit(SplitBase):
    _default_type_description: t.ClassVar[
        t.Optional[str]
    ] = "Percentage splits definition."
    _compact_form: t.ClassVar[t.Optional[str]] = "<fraction>[,<folder>]"

    fraction: t.Optional[float] = pyd.Field(
        ...,
        gt=0.0,
        le=1.0,
        description=(
            "Fraction of the dataset to keep. "
            "A None value takes all the samples left by the other splits."
        ),
    )

    def split_size(self, n_samples: int) -> t.Optional[int]:
        return int(n_samples * self.fraction) if self.fraction is not None else None

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, value):
        if isinstance(value, PercSplit):
            return value
        if isinstance(value, (str, bytes, float)):
            data = {}
            if isinstance(value, float):
                data["fraction"] = value
            else:
                sz, _, fld = str(value).partition(",")
                try:
                    data["fraction"] = float(sz)
                except ValueError:
                    # NB: leave size as-is if not valid, it will raise an error
                    data["fraction"] = (
                        None if sz.lower() in ("none", "null", "nul") else sz
                    )
                if fld:
                    data["output"] = pl_interfaces.OutputDatasetInterface(
                        folder=fld  # type: ignore
                    )
            value = data
        if isinstance(value, t.Mapping):
            return PercSplit(**value)
        raise ValueError("Invalid perc split definition.")


class AbsoluteSplit(SplitBase):
    _default_type_description: t.ClassVar[
        t.Optional[str]
    ] = "Absolute splits definition."
    _compact_form: t.ClassVar[t.Optional[str]] = "<length>[,<folder>]"

    length: t.Optional[pyd.PositiveInt] = pyd.Field(
        ...,
        description=(
            "Number of elements to keep. "
            "A None value takes all the samples left by the other splits."
        ),
    )

    def split_size(self, *args, **kwargs) -> t.Optional[int]:
        return self.length

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, value):
        if isinstance(value, AbsoluteSplit):
            return value
        if isinstance(value, (str, bytes, int)):
            data = {}
            if isinstance(value, int):
                data["length"] = value
            else:
                sz, _, fld = str(value).partition(",")
                try:
                    data["length"] = int(sz)
                except ValueError:
                    # NB: leave size as-is if not valid, it will raise an error
                    data["length"] = (
                        None if sz.lower() in ("none", "null", "nul") else sz
                    )

                if fld:
                    data["output"] = pl_interfaces.OutputDatasetInterface(
                        folder=fld  # type: ignore
                    )
            value = data
        if isinstance(value, t.Mapping):
            return AbsoluteSplit(**value)
        raise ValueError("Invalid absolute split definition.")


class Splits(pl_interfaces.PydanticFieldNoDefaultMixin):
    _default_type_description: t.ClassVar[t.Optional[str]] = "Splits definition."
    _compact_form: t.ClassVar[t.Optional[str]] = "<fraction|length>[,<folder>]"

    any_split = t.Union[
        PercSplit, AbsoluteSplit, t.Sequence[t.Union[PercSplit, AbsoluteSplit]]
    ]


class SplitCommand(PipelimeCommand, title="split"):
    """Split a dataset."""

    input: pl_interfaces.InputDatasetInterface = (
        pl_interfaces.InputDatasetInterface.pyd_field(
            alias="i", piper_port=PiperPortType.INPUT
        )
    )

    shuffle: t.Union[bool, pyd.PositiveInt] = pyd.Field(
        False,
        alias="shf",
        description=(
            "Shuffle the dataset before subsampling and splitting. "
            "Optionally specify the random seed."
        ),
    )
    subsample: pyd.PositiveInt = pyd.Field(
        1,
        alias="ss",
        description="Take 1-every-nth input sample. Applied after shuffling.",
    )

    splits: Splits.any_split = Splits.pyd_field(
        alias="s", piper_port=PiperPortType.OUTPUT
    )

    grabber: pl_interfaces.GrabberInterface = pl_interfaces.GrabberInterface.pyd_field(
        alias="g"
    )

    def run(self):
        reader = self.input.create_reader()
        if self.shuffle:
            reader = reader.shuffle(
                seed=self.shuffle if not isinstance(self.shuffle, bool) else None
            )
        if self.subsample != 1:
            reader = reader[:: self.subsample]
        input_length = len(reader)

        splits = self.splits if isinstance(self.splits, t.Sequence) else [self.splits]
        split_sizes = [s.split_size(input_length) for s in splits]
        none_idx = -1
        split_total = 0
        for i, s in enumerate(split_sizes):
            if s is None:
                if none_idx >= 0:
                    raise ValueError("Only one split size can be None.")
                none_idx = i
            else:
                split_total += s

        if split_total > input_length:
            raise ValueError(
                f"The sum of the split sizes ({split_total}) "
                f"is greater than the input length ({input_length})."
            )
        if none_idx >= 0:
            split_sizes[none_idx] = input_length - split_total

        split_start = 0
        for idx, split_length, split in zip(
            range(len(split_sizes)), split_sizes, splits
        ):
            split_stop = split_start + split_length  # type: ignore
            if split.output is not None:
                seq = reader[split_start:split_stop]
                seq = split.output.append_writer(seq)
                self.grabber.grab_all(
                    seq,
                    grab_context_manager=split.output.serialization_cm(),
                    keep_order=False,
                    parent_cmd=self,
                    track_message=(
                        f"Writing split {idx + 1}/{len(split_sizes)} "
                        f"({split_length} samples)"
                    ),
                )
            split_start = split_stop


class SplitByQueryCommand(PipelimeCommand, title="split-query"):
    """Split a dataset by query."""

    input: pl_interfaces.InputDatasetInterface = (
        pl_interfaces.InputDatasetInterface.pyd_field(
            alias="i", piper_port=PiperPortType.INPUT
        )
    )

    query: str = pyd.Field(
        ...,
        alias="q",
        description=("A query to match (cfr. https://github.com/cyberlis/dictquery)."),
    )

    output_selected: t.Optional[
        pl_interfaces.OutputDatasetInterface
    ] = pl_interfaces.OutputDatasetInterface.pyd_field(
        alias="os",
        is_required=False,
        description=(
            "Output dataset of sample selected by the query. "
            "Leave to None to not save to disk."
        ),
        piper_port=PiperPortType.OUTPUT,
    )

    output_discarded: t.Optional[
        pl_interfaces.OutputDatasetInterface
    ] = pl_interfaces.OutputDatasetInterface.pyd_field(
        alias="od",
        is_required=False,
        description=(
            "Output dataset of sample discarded by the query. "
            "Leave to None to not save to disk."
        ),
        piper_port=PiperPortType.OUTPUT,
    )

    grabber: pl_interfaces.GrabberInterface = pl_interfaces.GrabberInterface.pyd_field(
        alias="g"
    )

    def run(self):
        reader = self.input.create_reader()
        if self.output_selected is not None:
            self._filter(
                reader,
                self.output_selected,
                lambda x: x.match(self.query),
                "selected samples",
            )
        if self.output_discarded is not None:
            self._filter(
                reader,
                self.output_selected,
                lambda x: not x.match(self.query),
                "discarded samples",
            )

    def _filter(self, reader, output, fn, message):
        seq = reader.filter(fn)
        seq = output.append_writer(seq)
        self.grabber.grab_all(
            seq,
            grab_context_manager=output.serialization_cm(),
            keep_order=False,
            parent_cmd=self,
            track_message=f"Writing {message} ({len(seq)} samples)",
        )


class SplitByValueCommand(PipelimeCommand, title="split-value"):
    """Split a dataset into multiple sequences,
    one for each unique value of a given key."""

    input: pl_interfaces.InputDatasetInterface = (
        pl_interfaces.InputDatasetInterface.pyd_field(
            alias="i", piper_port=PiperPortType.INPUT
        )
    )

    key: str = pyd.Field(
        ...,
        alias="k",
        description=(
            "The key to use. A pydash-like dot notation "
            "can be used to access nested attributes."
        ),
    )

    output: pl_interfaces.OutputDatasetInterface = (
        pl_interfaces.OutputDatasetInterface.pyd_field(
            alias="o",
            description=(
                "Common options for the output sequences, "
                "which will be placed in subfolders."
            ),
            piper_port=PiperPortType.OUTPUT,
        )
    )

    grabber: pl_interfaces.GrabberInterface = pl_interfaces.GrabberInterface.pyd_field(
        alias="g"
    )

    def run(self):
        import uuid

        from pipelime.sequences import Sample

        class WorkerHelper:
            def __init__(self, idx_key, value_key):
                self._idx_key = idx_key
                self._value_key = value_key
                self._groups = {}

            def _value_to_str(self, value):
                import numpy as np

                if isinstance(value, np.ndarray):
                    if value.size == 1:
                        value = (
                            int(value)
                            if np.issubdtype(value.dtype, np.integer)
                            else float(value)
                        )
                    else:
                        value = tuple(value)
                return (
                    str(value)
                    .replace(", ", "„")
                    .replace(",", "„")
                    .replace(" ", "_")
                    .replace("/", "-")
                    .replace("\\", "-")
                    .replace("<", "‹")
                    .replace(">", "›")
                    .replace(":", "⁚")
                    .replace('"', "'")
                    .replace("|", "¦")
                    .replace("?", "⁇")
                    .replace("*", "⁎")
                )

            def __call__(self, x: Sample):
                value = x.deep_get(self._value_key)
                if value is not None:
                    value = self._value_to_str(value)
                    self._groups.setdefault(value, []).append(
                        int(x[self._idx_key]())  # type: ignore
                    )

        reader = self.input.create_reader()

        unique_idx_key = uuid.uuid1().hex
        worker = WorkerHelper(unique_idx_key, self.key)
        self.grabber.grab_all(
            reader.enumerate(idx_key=unique_idx_key),
            keep_order=True,
            parent_cmd=self,
            track_message="Gathering unique values",
            sample_fn=worker,
        )

        for idx, (group_val, group_idxs) in enumerate(worker._groups.items()):
            split_name = f"{self.key}={group_val}"
            split_output = self.output.copy(
                update={"folder": self.output.folder / split_name}
            )

            split_name = f"{split_name} " if len(split_name) < 20 else ""
            split_seq = split_output.append_writer(reader.select(group_idxs))
            self.grabber.grab_all(
                split_seq,
                grab_context_manager=split_output.serialization_cm(),
                keep_order=False,
                parent_cmd=self,
                track_message=(
                    f"Writing split {idx + 1}/{len(worker._groups)} "
                    f"{split_name}({len(split_seq)} samples)"
                ),
            )



import os
from contextlib import redirect_stdout
from pathlib import Path
import rich

from pipelime.cli import print_commands_ops_stages_list

os.environ["COLUMNS"] = "130"

base_path = Path(__file__).parent.resolve() / "_static/generated"
base_path.mkdir(parents=True, exist_ok=True)

rich.reconfigure(record=True)

with open(os.devnull, "w", encoding="utf-8") as f, redirect_stdout(f):
    print_commands_ops_stages_list(
        True, show_cmds=True, show_ops=False, show_stages=False
    )

with (base_path / "pl_cmds.html").open("w", encoding="utf-8") as f:
    f.write(rich.get_console().export_html())

with open(os.devnull, "w", encoding="utf-8") as f, redirect_stdout(f):
    print_commands_ops_stages_list(
        True, show_cmds=False, show_ops=False, show_stages=True
    )

with (base_path / "pl_stgs.html").open("w", encoding="utf-8") as f:
    f.write(rich.get_console().export_html())

with open(os.devnull, "w", encoding="utf-8") as f, redirect_stdout(f):
    print_commands_ops_stages_list(
        True, show_cmds=False, show_ops=True, show_stages=False
    )

with (base_path / "pl_ops.html").open("w", encoding="utf-8") as f:
    f.write(rich.get_console().export_html())



from pipelime.sequences import Sample
from pipelime.items import UnknownItem


class TestBaseStages:
    def _check_target_pred(self, target: Sample, pred: Sample):
        assert all(
            k in target and type(target[k]) == type(v) and target[k]() == v()
            for k, v in pred.items()
        )
        assert all(
            k in pred and type(pred[k]) == type(v) and pred[k]() == v()
            for k, v in target.items()
        )

    def test_identity(self):
        from pipelime.stages import StageIdentity

        stage = StageIdentity()
        sample = Sample({"a": UnknownItem()})
        assert stage(sample) is sample

    def test_lambda(self):
        from pipelime.stages import StageLambda

        stage = StageLambda(
            func=(lambda x: x.set_item("a", UnknownItem(x["a"]() * 2)))  # type: ignore
        )
        self._check_target_pred(
            Sample({"a": UnknownItem(20)}), stage(Sample({"a": UnknownItem(10)}))
        )

    def test_compose(self):
        from pipelime.stages import StageCompose, StageLambda

        stage = StageCompose(
            [
                StageLambda(
                    func=(
                        lambda x: x.set_item("a", UnknownItem(x["a"]() * 2))  # type: ignore
                    )
                ),
                StageLambda(
                    func=(
                        lambda x: x.set_item("a", UnknownItem(x["a"]() + 5))  # type: ignore
                    )
                ),
            ]
        )
        self._check_target_pred(
            Sample({"a": UnknownItem(25)}), stage(Sample({"a": UnknownItem(10)}))
        )



from __future__ import annotations
from abc import abstractmethod
import itertools
import typing as t
import pydantic as pyd
from loguru import logger

from pipelime.sequences.sample import Sample


class SamplesSequenceBase(t.Sequence[Sample]):
    @abstractmethod
    def size(self) -> int:
        pass

    @abstractmethod
    def get_sample(self, idx: int) -> Sample:
        pass

    def __len__(self) -> int:
        return self.size()

    def is_normalized(self, max_items=-1) -> bool:
        """Checks if all samples have the same keys.

        :param max_items: limits to the first `max_items`, defaults to -1
        :type max_items: int, optional
        :return: True if all samples have the same keys
        :rtype: bool
        """
        max_items = len(self) if max_items < 0 else min(max_items, len(self))
        if max_items < 2:
            return True
        it = itertools.islice(self, max_items)
        key_ref = set(next(it).keys())
        for s in it:
            if key_ref != set(s.keys()):
                return False
        return True

    def best_zfill(self) -> int:
        """Computes the best zfill for integer indexing.

        :return: zfill values (maximum number of digits based on current size)
        :rtype: int
        """
        return len(str(len(self) - 1))


class SamplesSequence(
    SamplesSequenceBase, pyd.BaseModel, extra="forbid", copy_on_model_validation="none"
):
    """A generic sequence of samples. Subclasses should implement `size(self) -> int`
    and `get_sample(self, idx: int) -> Sample`.

    The list of all available pipes and sources can be retrieved
    through `pipelime list-ops`.

    NB: when defining a pipe, the `source` sample sequence must be bound to a pydantic
    Field with `pipe_source=True`.
    """

    _sources: t.ClassVar[t.Dict[str, t.Type[SamplesSequence]]] = {}
    _pipes: t.ClassVar[t.Dict[str, t.Type[SamplesSequence]]] = {}
    _operator_path: t.ClassVar[str] = ""

    @t.overload
    def __getitem__(self, idx: int) -> Sample:
        ...

    @t.overload
    def __getitem__(self, idx: slice) -> SamplesSequence:
        ...

    def __getitem__(self, idx: t.Union[int, slice]) -> t.Union[Sample, SamplesSequence]:
        return (
            self.slice(start=idx.start, stop=idx.stop, step=idx.step)
            if isinstance(idx, slice)
            else self.get_sample(idx if idx >= 0 else len(self) + idx)
        )

    def __iter__(self) -> t.Iterator[Sample]:
        _local_idx = 0
        while _local_idx < len(self):
            yield self[_local_idx]
            _local_idx += 1

    def __add__(self, other: SamplesSequence) -> SamplesSequence:
        return self.cat(other)

    @classmethod
    def name(cls) -> str:
        if cls.__config__.title:
            return cls.__config__.title
        return cls.__name__

    def direct_access(self) -> t.Sequence[t.Mapping[str, t.Any]]:
        """Returns a sequence of key-to-value mappings,
        with no intermediate Sample and Item classes.
        """
        from pipelime.sequences.direct_access import DirectAccessSequence

        return DirectAccessSequence(self)

    def torch_dataset(self) -> "torch.utils.data.Dataset":  # type: ignore # noqa: E602,F821
        """Returns a torch.utils.data.Dataset interface of this samples sequence."""
        from pipelime.sequences.torch import TorchDataset

        return TorchDataset(self)

    def batch(self, batch_size: int, drop_last: bool = False, fill: Sample = Sample()):
        """Returns a zip-like object to get batches of samples. If the number of
        samples is not a multiple of batch_size and `drop_last` is False,
        the last batch will be filled with `fill`.
        """
        src_iters = [iter(self)] * batch_size
        if drop_last:
            return zip(*src_iters)
        return itertools.zip_longest(*src_iters, fillvalue=fill)

    def apply(
        self,
        *,
        num_workers: int = 0,
        prefetch: int = 2,
        track_fn: t.Optional[t.Callable[[t.Iterable], t.Iterable]] = None,
    ):
        """Goes through all the samples of the sequence, optionally using multiple
        processes and returns a new sequence holding the processed samples.
        Also, a `track_fn` can be defined, eg, to show the progress.

        :param num_workers: The number of processes to spawn. If negative,
            the number of (logical) cpu cores is used, defaults to 0
        :type num_workers: int, optional
        :param prefetch: The number of samples loaded in advanced by each worker,
            defaults to 2
        :type prefetch: int, optional
        :param sample_fn: a callable to run on each sample, defaults to None
        :type sample_fn: t.Optional[t.Callable[[Sample], None]], optional
        :param track_fn: a callable to track the progress, defaults to None
        :type track_fn: t.Optional[t.Callable[[t.Iterable], t.Iterable]], optional
        """
        samples: t.List[Sample] = []

        def _store_sample(x: Sample, idx: int):
            if len(samples) <= idx:
                samples.extend([Sample() for _ in range(idx - len(samples) + 1)])
            samples[idx] = x

        self.run(
            num_workers=num_workers,
            prefetch=prefetch,
            keep_order=False,
            sample_fn=_store_sample,
            track_fn=track_fn,
        )

        return SamplesSequence.from_list(samples)

    def run(
        self,
        *,
        num_workers: int = 0,
        prefetch: int = 2,
        keep_order: bool = False,
        sample_fn: t.Union[
            t.Callable[[Sample], None], t.Callable[[Sample, int], None], None
        ] = None,
        track_fn: t.Optional[t.Callable[[t.Iterable], t.Iterable]] = None,
    ):
        """Goes through all the samples of the sequence, optionally using multiple
        processes and applying `sample_fn` to each sample. Also, a `track_fn` can be
        defined, eg, to show the progress.

        :param num_workers: The number of processes to spawn. If negative,
            the number of (logical) cpu cores is used, defaults to 0
        :type num_workers: int, optional
        :param prefetch: The number of samples loaded in advanced by each worker,
            defaults to 2
        :type prefetch: int, optional
        :param keep_order: Whether to retrieve the samples in the original order,
            defaults to False
        :type keep_order: bool, optional
        :param sample_fn: a callable to run on each sample, defaults to None
        :type sample_fn: t.Optional[t.Callable[[Sample], None]], optional
        :param track_fn: a callable to track the progress, defaults to None
        :type track_fn: t.Optional[t.Callable[[t.Iterable], t.Iterable]], optional
        """
        from pipelime.sequences import Grabber, grab_all

        grabber = Grabber(
            num_workers=num_workers, prefetch=prefetch, keep_order=keep_order
        )
        grab_all(grabber, self, sample_fn=sample_fn, track_fn=track_fn)

    def to_pipe(
        self, recursive: bool = True, objs_to_str: bool = True
    ) -> t.List[t.Dict[str, t.Any]]:
        """Serializes this sequence to a pipe list. You can then pass this list to
        `pipelime.sequences.build_pipe` to reconstruct the sequence.
        NB: nested sequences are recursively serialized only if `recursive` is True,
        while other objects are not. Consider to use `pipelime.choixe` features to
        fully (de)-serialized them to yaml/json.

        :param recursive: if True nested sequences are recursively serialized,
            defaults to True.
        :type recursive: bool, optional.
        :param objs_to_str: if True objects are converted to string.
        :type objs_to_str: bool, optional.
        :raises ValueError: if a field is tagged as `pipe_source` but it is not
            a SamplesSequence.
        :return: the serialized pipe list.
        :rtype: t.List[t.Dict[str, t.Any]]
        """
        source_list = []
        arg_dict = {}
        for field_name, model_field in self.__fields__.items():
            field_value = getattr(self, field_name)
            field_alias = model_field.alias
            if model_field.field_info.extra.get("pipe_source", False):
                if not isinstance(field_value, SamplesSequence):
                    raise ValueError(
                        f"{field_alias} is tagged as `pipe_source`, "
                        "but it is not a SamplesSequence instance."
                    )
                source_list = field_value.to_pipe(
                    recursive=recursive, objs_to_str=objs_to_str
                )
            else:
                # NB: do not unfold sub-pydantic models, since it may not be
                # straightforward to de-serialize them when subclasses are used
                if recursive:
                    if isinstance(field_value, SamplesSequence):
                        field_value = field_value.to_pipe(
                            recursive=recursive, objs_to_str=objs_to_str
                        )
                    elif isinstance(field_value, pyd.BaseModel):
                        field_value = field_value.dict()
                arg_dict[field_alias] = (
                    (
                        list(field_value)
                        if isinstance(field_value, tuple)
                        else field_value
                    )
                    if not objs_to_str
                    or isinstance(
                        field_value,
                        (str, bytes, int, float, bool, t.Mapping, t.Sequence),
                    )
                    or field_value is None
                    else str(field_value)
                )
        return source_list + [{self._operator_path: arg_dict}]

    def __str__(self) -> str:
        return repr(self)

    ###########################################################################
    # FUNCTION STUBS FOR TYPE CHECKING AND AUTO-COMPLETION
    ###########################################################################

    @staticmethod
    def from_callable(
        *,
        generator_fn: t.Callable[[int], Sample],
        length: t.Union[int, t.Callable[[], int]],
    ) -> SamplesSequence:
        """A SamplesSequence calling a user-defined generator to get the samples.
        Run `pipelime help from_callable` to read the complete documentation.
        """
        ...

    @staticmethod
    def from_list(samples: t.Sequence[Sample]) -> SamplesSequence:
        """A SamplesSequence from a list of Samples.
        Run `pipelime help from_list` to read the complete documentation.
        """
        ...

    @staticmethod
    def from_underfolder(
        folder: "pathlib.Path",  # type: ignore # noqa: E602,F821
        *,
        merge_root_items: bool = True,
        must_exist: bool = True,
        watch: bool = False,
    ) -> SamplesSequence:
        """A SamplesSequence loading data from an Underfolder dataset.
        Run `pipelime help from_underfolder` to read the complete documentation.
        """
        ...

    @staticmethod
    def toy_dataset(
        length: int,
        *,
        with_images: bool = True,
        with_masks: bool = True,
        with_instances: bool = True,
        with_objects: bool = True,
        with_bboxes: bool = True,
        with_kpts: bool = True,
        image_size: int = 256,
        key_format: str = "*",
        max_labels: int = 5,
        objects_range: t.Tuple[int, int] = (1, 5),
    ) -> SamplesSequence:
        """A fake sequence of generated samples.
        Run `pipelime help toy_dataset` to read the complete documentation.
        """
        ...

    def map(
        self,
        stage: t.Union[
            "pipelime.stages.SampleStage",  # type: ignore # noqa: E602,F821
            t.Mapping[str, t.Optional[t.Mapping[str, t.Any]]],
        ],
    ) -> SamplesSequence:
        """Applies a stage on all samples.
        Run `pipelime help map` to read the complete documentation.
        """
        ...

    def zip(self, to_zip: SamplesSequence, *, key_format: str = "*") -> SamplesSequence:
        """Zips two Sequences by merging each Sample.
        Run `pipelime help zip` to read the complete documentation.
        """
        ...

    def cat(self, to_cat: SamplesSequence) -> SamplesSequence:
        """Concatenates two SamplesSequences.
        Run `pipelime help cat` to read the complete documentation.
        """
        ...

    def filter(self, filter_fn: t.Callable[[Sample], bool]) -> SamplesSequence:
        """A filtered view of a SamplesSequence.
        Run `pipelime help filter` to read the complete documentation.
        """
        ...

    def sort(self, key_fn: t.Callable[[Sample], t.Any]) -> SamplesSequence:
        """A sorted view of an input SamplesSequence.
        Run `pipelime help sort` to read the complete documentation.
        """
        ...

    def slice(
        self,
        *,
        start: t.Optional[int] = None,
        stop: t.Optional[int] = None,
        step: t.Optional[int] = None,
    ) -> SamplesSequence:
        """Functional version of the slice operator `self[start_idx:end_idx:step]`.
        Run `pipelime help slice` to read the complete documentation.
        """
        ...

    def select(
        self, indexes: t.Sequence[int], *, negate: bool = False
    ) -> SamplesSequence:
        """Given a list of indexes, extracts the corresponding samples
        from the input SamplesSequence. The index sequence is not automatically sorted.
        Run `pipelime help select` to read the complete documentation.
        """
        ...

    def shuffle(self, *, seed: t.Optional[int] = None) -> SamplesSequence:
        """Shuffles samples in the input SamplesSequence.
        Run `pipelime help shuffle` to read the complete documentation.
        """
        ...

    def enumerate(
        self,
        *,
        idx_key: str = "~idx",
        item_cls: str = "pipelime.items.TxtNumpyItem",
    ) -> SamplesSequence:
        """Add a new index item to each Sample in the input SamplesSequence.
        Run `pipelime help enumerate` to read the complete documentation.
        """
        ...

    def repeat(self, count: int) -> SamplesSequence:
        """Repeat this sequence so each sample is seen multiple times.
        Run `pipelime help repeat` to read the complete documentation.
        """
        ...

    def cache(
        self,
        cache_folder: t.Optional["pathlib.Path"] = None,  # type: ignore # noqa: E602,F821
        *,
        reuse_cache: bool = False,
    ) -> SamplesSequence:
        """Cache the input Samples the first time they are accessed.
        Run `pipelime help cache` to read the complete documentation.
        """
        ...

    def data_cache(
        self, *items: t.Union[t.Type["pipelime.items.Item"], str]  # type: ignore # noqa: E602,F821
    ) -> SamplesSequence:
        """Enables item data caching on previous pipeline steps."""
        ...

    def no_data_cache(
        self, *items: t.Union[t.Type["pipelime.items.Item"], str]  # type: ignore # noqa: E602,F821
    ) -> SamplesSequence:
        """Disables item data caching on previous pipeline steps."""
        ...

    def to_underfolder(
        self,
        folder: "pathlib.Path",  # type: ignore # noqa: E602,F821
        *,
        zfill: t.Optional[int] = None,
        key_serialization_mode: t.Optional[
            t.Mapping[
                str, t.Union["pipelime.items.SerializationMode", str]  # type: ignore # noqa: E602,F821
            ]
        ] = None,
        exists_ok: bool = False,
    ) -> SamplesSequence:
        """Writes samples to an underfolder dataset while iterating over them.
        Run `pipelime help to_underfolder` to read the complete documentation.
        """
        ...

    def validate_samples(
        self,
        *,
        sample_schema: t.Type[pyd.BaseModel],
        lazy: bool = False,
        max_samples: int = 1,
    ) -> SamplesSequence:
        """Validates the source sequence against a schema.
        Run `pipelime help validate_samples` to read the complete documentation.
        """
        ...

    ###########################################################################
    # STUBS END
    ###########################################################################


def _add_operator_path(cls: t.Type[SamplesSequence]) -> t.Type[SamplesSequence]:
    from pathlib import Path
    import inspect

    if cls.__module__.startswith("pipelime"):
        cls._operator_path = cls.name()
    else:
        try:
            source_path = inspect.getfile(cls)
            module_path = Path(source_path).resolve().as_posix()
        except TypeError:
            # this happens when the class does not come from a file
            module_path = "__main__"

        if (cls.__module__.replace(".", "/") + ".py") in module_path:
            module_path = cls.__module__
        cls._operator_path = module_path + ":" + cls.name()
    return cls


def source_sequence(cls: t.Type[SamplesSequence]) -> t.Type[SamplesSequence]:
    if cls.name() in SamplesSequence._sources or cls.name() in SamplesSequence._pipes:
        logger.warning(f"Function {cls.name()} has been already registered.")

    setattr(SamplesSequence, cls.name(), cls)
    SamplesSequence._sources[cls.name()] = cls
    cls = _add_operator_path(cls)
    return cls


def piped_sequence(cls: t.Type[SamplesSequence]) -> t.Type[SamplesSequence]:
    if cls.name() in SamplesSequence._sources or cls.name() in SamplesSequence._pipes:
        logger.warning(f"Function {cls.name()} has been already registered.")

    prms_source_name = None
    for mfield in cls.__fields__.values():
        if mfield.field_info.extra.get("pipe_source", False):
            if prms_source_name is not None:
                raise ValueError(
                    f"More than one field has `pipe_source=True` in {cls.__name__}."
                )
            prms_source_name = mfield.alias
    if prms_source_name is None:
        raise ValueError(
            f"{cls.__name__} is tagged as `piped`, but no field has `pipe_source=True`."
        )

    def _helper(self, *args, **kwargs):
        return cls(*args, **{prms_source_name: self}, **kwargs)

    setattr(SamplesSequence, cls.name(), _helper)
    SamplesSequence._pipes[cls.name()] = cls
    cls = _add_operator_path(cls)
    return cls



from __future__ import annotations
import typing as t
from pathlib import Path
from urllib.parse import ParseResult

import pydantic as pyd

from pipelime.utils.pydantic_types import ItemType


class PydanticFieldMixinBase:
    # override in derived clasess
    _default_type_description: t.ClassVar[t.Optional[str]] = None
    _compact_form: t.ClassVar[t.Optional[str]] = None

    @classmethod
    def _description(cls, user_desc: t.Optional[str]) -> t.Optional[str]:
        from pipelime.cli.pretty_print import _short_line

        desc_list = []
        if user_desc is None:
            user_desc = cls._default_type_description
            if user_desc is not None:
                desc_list.append(user_desc)
            elif cls._compact_form is None:
                return None
        if cls._compact_form is not None:
            desc_list.append(f"{_short_line()} Compact form: `{cls._compact_form}`")
        return "\n".join(desc_list)


class PydanticFieldWithDefaultMixin(PydanticFieldMixinBase):
    @classmethod
    def pyd_field(cls, *, description: t.Optional[str] = None, **kwargs):
        return pyd.Field(
            default_factory=cls,  # type: ignore
            description=cls._description(description),  # type: ignore
            **kwargs,
        )


class PydanticFieldNoDefaultMixin(PydanticFieldMixinBase):
    @classmethod
    def pyd_field(
        cls, *, is_required: bool = True, description: t.Optional[str] = None, **kwargs
    ):
        return pyd.Field(
            ... if is_required else None,
            description=cls._description(description),  # type: ignore
            **kwargs,
        )


class GrabberInterface(PydanticFieldWithDefaultMixin, pyd.BaseModel, extra="forbid"):
    """Multiprocessing grabbing options.

    Examples:
        How to use it in your command::

            class EasyCommand(PipelimeCommand, title="easy"):
                input: InputDatasetInterface = InputDatasetInterface.pyd_field(
                    alias="i", piper_port=PiperPortType.INPUT
                )
                output: OutputDatasetInterface = OutputDatasetInterface.pyd_field(
                        alias="o", piper_port=PiperPortType.OUTPUT
                )
                grabber: GrabberInterface = GrabberInterface.pyd_field(
                    alias="g"
                )

                def run(self):
                    seq = self.input.create_reader()
                    seq = self.output.append_writer(seq)
                    self.grabber.grab_all(
                        seq,
                        grab_context_manager=self.output.serialization_cm(),
                        keep_order=False,
                        parent_cmd=self,
                        track_message=f"Executing ({len(seq)} samples)",
                    )
    """

    _default_type_description: t.ClassVar[t.Optional[str]] = "Grabber options."
    _compact_form: t.ClassVar[t.Optional[str]] = "<num_workers>[,<prefetch>]"

    num_workers: int = pyd.Field(
        0,
        description=(
            "The number of processes to spawn. If negative, "
            "the number of (logical) cpu cores is used."
        ),
    )
    prefetch: pyd.PositiveInt = pyd.Field(
        2, description="The number of samples loaded in advanced by each worker."
    )

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, value):
        if isinstance(value, GrabberInterface):
            return value

        if isinstance(value, (str, bytes, int)):
            data = {}
            if isinstance(value, int):
                data["num_workers"] = value
            else:
                wrks, _, pf = str(value).partition(",")
                data["num_workers"] = int(wrks)
                if pf:
                    data["prefetch"] = int(pf)
            value = data

        if isinstance(value, t.Mapping):
            return GrabberInterface(**value)
        raise ValueError("Invalid grabber definition.")

    def grab_all(
        self,
        sequence,
        *,
        keep_order: bool = False,
        parent_cmd=None,
        track_message: str = "",
        sample_fn=None,
        size: t.Optional[int] = None,
        grab_context_manager: t.Optional[t.ContextManager] = None,
    ):
        """Runs the grabber on a sequence.
        NB: `sample_fn` always runs on the main process and may take just the sample
        or the sample and its index.
        NB: `grab_context_manager.__enter__` will be used as `worker_init_fn`,
        please use `GrabberInterface.grab_all_wrk_init` if you want to specify
        your `worker_init_fn`.

        Args:
            sequence: the sequence to grab, usually a SamplesSequence.
            keep_order (bool, optional): if True, `sample_fn` will always receive the
                sample in the correct order. Defaults to False.
            parent_cmd (_type_, optional): the pipelime command running the grabber is
                needed to correctly setup the progress bar. Defaults to None.
            track_message (str, optional): a message shown next to the progress bar.
                Defaults to "".
            sample_fn (_type_, optional): an optional function to run on each grabbed
                element, usually a Sample. The signature may be (elem) or (elem, index).
                Defaults to None.
            size (t.Optional[int], optional): the size of the sequence. If not given,
                `len(sequence)` is evaluated. Defaults to None.
            grab_context_manager (ContextManager, optional): a context manager wrapping
                the whole grabbing operation on the main process. Also,
                `grab_context_manager.__enter__` will be used as `worker_init_fn`.
                Defaults to None.
        """
        from copy import deepcopy

        return self.grab_all_wrk_init(
            sequence=sequence,
            keep_order=keep_order,
            parent_cmd=parent_cmd,
            track_message=track_message,
            sample_fn=sample_fn,
            size=size,
            grab_context_manager=grab_context_manager,
            worker_init_fn=None
            if grab_context_manager is None
            else deepcopy(grab_context_manager).__enter__,
        )

    def grab_all_wrk_init(
        self,
        sequence,
        *,
        keep_order: bool = False,
        parent_cmd=None,
        track_message: str = "",
        sample_fn=None,
        size: t.Optional[int] = None,
        grab_context_manager: t.Optional[t.ContextManager] = None,
        worker_init_fn: t.Union[
            t.Callable, t.Tuple[t.Callable, t.Sequence], None
        ] = None,
    ):
        """Runs the grabber on a sequence.
        NB: `sample_fn` always runs on the main process and may take just the sample
        or the sample and its index.

        Args:
            sequence: the sequence to grab, usually a SamplesSequence.
            keep_order (bool, optional): if True, `sample_fn` will always receive the
                sample in the correct order. Defaults to False.
            parent_cmd (_type_, optional): the pipelime command running the grabber is
                needed to correctly setup the progress bar. Defaults to None.
            track_message (str, optional): a message shown next to the progress bar.
                Defaults to "".
            sample_fn (_type_, optional): an optional function to run on each grabbed
                element, usually a Sample. The signature may be (elem) or (elem, index).
                Defaults to None.
            size (t.Optional[int], optional): the size of the sequence. If not given,
                `len(sequence)` is evaluated. Defaults to None.
            grab_context_manager (ContextManager, optional): a context manager wrapping
                the whole grabbing operation on the main process. Defaults to None.
            worker_init_fn (optional): a callable or a tuple (callable, args) to run
                before starting the grabbing loop on each worker. If no process is
                spawn, it will be run on the main process. Defaults to None.
        """
        from pipelime.sequences import Grabber, grab_all

        grabber = Grabber(
            num_workers=self.num_workers, prefetch=self.prefetch, keep_order=keep_order
        )
        track_fn = (
            None
            if parent_cmd is None
            else (
                lambda x: parent_cmd.track(
                    x,
                    size=len(sequence) if size is None else size,
                    message=track_message,
                )
            )
        )
        grab_all(
            grabber,
            sequence,
            track_fn=track_fn,
            sample_fn=sample_fn,
            size=size,
            grab_context_manager=grab_context_manager,
            worker_init_fn=worker_init_fn,
        )


class ItemValidationModel(
    pyd.BaseModel, extra="forbid", copy_on_model_validation="none"
):
    """Item schema validation."""

    class_path: ItemType = pyd.Field(
        ...,
        description=(
            "The item class path. The default package `pipelime.item` can be omitted"
        ),
    )
    is_optional: bool = pyd.Field(
        True, description="Whether the item is required or optional."
    )
    is_shared: bool = pyd.Field(False, description="Whether the item is shared or not.")
    validator_: t.Optional[str] = pyd.Field(
        None,
        description=(
            "A class path to a callable accepting the item value and either returning "
            "a validated value or raising an exception in case of error."
        ),
        alias="validator",
    )

    _validator_callable = pyd.PrivateAttr()

    def __init__(self, **data):
        from pipelime.choixe.utils.imports import import_symbol

        super().__init__(**data)

        def _identity(v):
            return v

        self._validator_callable = (
            import_symbol(self.validator_) if self.validator_ else _identity
        )

    def make_field(self, key_name: str):
        return (
            self.class_path.itype,
            pyd.Field(default_factory=self.class_path.itype, alias=key_name)
            if self.is_optional
            else pyd.Field(..., alias=key_name),
        )

    def make_validator_method(self, field_name: str):
        import uuid

        # we need random names and dynamic function creation
        # to avoid reusing the same function name for validators
        # (yes, pydantic is really pedantic...)
        rnd_name = uuid.uuid1().hex

        _validator_wrapper = (
            "def validate_{}_fn(cls, v):\n".format(rnd_name)
            + "    if v.is_shared != {}:\n".format(self.is_shared)
            + "        raise ValueError(\n"
            + "            'Item must{}be shared.'\n".format(
                " not " if not self.is_shared else " "
            )
            + "        )\n"
            + "    return user_validator_{}(v)\n".format(rnd_name)
        )

        local_scope = {
            **globals(),
            f"user_validator_{rnd_name}": self._validator_callable,
        }
        exec(_validator_wrapper, local_scope)
        fn_helper = local_scope[f"validate_{rnd_name}_fn"]
        return pyd.validator(field_name)(fn_helper)


class SampleValidationInterface(
    pyd.BaseModel, extra="forbid", copy_on_model_validation="none"
):
    """Sample schema validation."""

    sample_schema: t.Union[str, t.Mapping[str, ItemValidationModel]] = pyd.Field(
        ...,
        description=(
            "The sample schema to validate, ie, a mapping from sample keys to expected "
            "item types.\nThe schema can be a class path to a pydantic model where "
            "fields' names are the sample keys, while fields' values are the item "
            "types. Otherwise, an explicit `key-name: ItemValidationModel` mapping "
            "must be provided."
        ),
    )
    ignore_extra_keys: bool = pyd.Field(
        True,
        description=(
            "When `sample_schema` is an explicit mapping, if `ignore_extra_keys` is "
            "True, unexpected keys are ignored. Otherwise an error is raised."
        ),
    )
    lazy: bool = pyd.Field(
        True, description="If True, samples will be validated only when accessed."
    )
    max_samples: int = pyd.Field(
        1,
        description=(
            "When the validation is NOT lazy, "
            "only the slice `[0:max_samples]` is checked. "
            "Set to 0 to check all the samples."
        ),
    )

    _schema_model: t.Type[pyd.BaseModel] = pyd.PrivateAttr()

    def _import_schema(self, schema_path: str):
        from pipelime.choixe.utils.imports import import_symbol

        imported_schema = import_symbol(schema_path)
        if not issubclass(imported_schema, pyd.BaseModel):
            raise ValueError(f"`{schema_path}` is not a pydantic model.")
        return imported_schema

    def _make_schema(self, schema_def: t.Mapping[str, ItemValidationModel]):
        class Config(pyd.BaseConfig):
            arbitrary_types_allowed = True
            extra = pyd.Extra.ignore if self.ignore_extra_keys else pyd.Extra.forbid

        def _safe_name(k):
            return f"{k}___"

        _item_map = {_safe_name(k): v.make_field(k) for k, v in schema_def.items()}
        _validators = {
            f"validate_{k}": v.make_validator_method(_safe_name(k))
            for k, v in schema_def.items()
        }

        return pyd.create_model(
            "SampleSchema",
            __config__=Config,
            __validators__=_validators,
            **_item_map,
        )

    def __init__(self, **data):
        super().__init__(**data)
        self._schema_model = (
            self._import_schema(self.sample_schema)
            if isinstance(self.sample_schema, str)
            else self._make_schema(self.sample_schema)
        )

    def append_validator(self, sequence):
        return sequence.validate_samples(
            sample_schema=self._schema_model,
            lazy=self.lazy,
            max_samples=self.max_samples,
        )

    def as_pipe(self):
        return {
            "validate_samples": {
                "sample_schema": self._schema_model,
                "lazy": self.lazy,
                "max_samples": self.max_samples,
            }
        }


class InputDatasetInterface(
    PydanticFieldNoDefaultMixin,
    pyd.BaseModel,
    extra="forbid",
    copy_on_model_validation="none",
):
    """Input dataset options.

    Examples:
        How to use it in your command::

            class EasyCommand(PipelimeCommand, title="easy"):
                input: InputDatasetInterface = InputDatasetInterface.pyd_field(
                    alias="i", piper_port=PiperPortType.INPUT
                )
                output: OutputDatasetInterface = OutputDatasetInterface.pyd_field(
                        alias="o", piper_port=PiperPortType.OUTPUT
                )
                grabber: GrabberInterface = GrabberInterface.pyd_field(
                    alias="g"
                )

                def run(self):
                    seq = self.input.create_reader()
                    seq = self.output.append_writer(seq)
                    self.grabber.grab_all(
                        seq,
                        grab_context_manager=self.output.serialization_cm(),
                        keep_order=False,
                        parent_cmd=self,
                        track_message=f"Executing ({len(seq)} samples)",
                    )
    """

    _default_type_description: t.ClassVar[t.Optional[str]] = "The input dataset."
    _compact_form: t.ClassVar[t.Optional[str]] = "<folder>[,<skip_empty>]"

    folder: Path = pyd.Field(..., description="Dataset root folder.")
    merge_root_items: bool = pyd.Field(
        True,
        description=(
            "Adds root items as shared items "
            "to each sample (sample values take precedence)."
        ),
    )
    skip_empty: bool = pyd.Field(False, description="Filter out empty samples.")
    schema_: t.Optional[SampleValidationInterface] = pyd.Field(
        None, alias="schema", description="Sample schema validation."
    )

    @pyd.validator("folder")
    def resolve_folder(cls, v: Path):
        # see https://bugs.python.org/issue38671
        return v.resolve().absolute()

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, value):
        if isinstance(value, InputDatasetInterface):
            return value

        if isinstance(value, (str, bytes)):
            fld, _, sk_emp = str(value).partition(",")
            data: t.Mapping[str, t.Any] = {"folder": fld}
            if sk_emp:
                # NB: the value is left as-is when it is not True nor False
                # to raise an error on validation
                data["skip_empty"] = (  # type: ignore
                    True
                    if sk_emp.lower() == "true"
                    else (False if sk_emp.lower() == "false" else sk_emp)
                )
            value = data

        if isinstance(value, t.Mapping):
            return InputDatasetInterface(**value)
        raise ValueError("Invalid input dataset definition.")

    def create_reader(self):
        from pipelime.sequences import SamplesSequence

        reader = SamplesSequence.from_underfolder(  # type: ignore
            folder=self.folder, merge_root_items=self.merge_root_items, must_exist=True
        )
        if self.skip_empty:
            reader = reader.filter(lambda x: not all(i.is_shared for i in x.values()))
        if self.schema_ is not None:
            reader = self.schema_.append_validator(reader)
        return reader

    def __repr__(self) -> str:
        return self.__piper_repr__()

    def __piper_repr__(self) -> str:
        return self.folder.as_posix()


class SerializationModeInterface(
    pyd.BaseModel,
    extra="forbid",
    copy_on_model_validation="none",
    underscore_attrs_are_private=True,
):
    """Serialization modes for items and keys."""

    override: t.Mapping[str, t.Optional[t.Union[str, t.Sequence[str]]]] = pyd.Field(
        default_factory=dict,
        description=(
            "Serialization modes overridden for specific item types, "
            "eg, `{CREATE_NEW_FILE: [ImageItem, my.package.MyItem, "
            "my/module.py:OtherItem]}`. A Null value applies to all items."
        ),
    )
    disable: t.Mapping[str, t.Union[str, t.Sequence[str]]] = pyd.Field(
        default_factory=dict,
        description=(
            "Serialization modes disabled for specific item types, "
            "eg, `{ImageItem: HARD_LINK, my.package.MyItem: [SYM_LINK, DEEP_COPY]}`. "
            "The special key `_` applies to all items."
        ),
    )
    keys: t.Mapping[str, str] = pyd.Field(
        default_factory=dict,
        description=(
            "Serialization modes overridden for specific sample keys, "
            "eg, `{image: CREATE_NEW_FILE}`."
        ),
    )

    _overridden_modes_cms: t.List[t.ContextManager]
    _disabled_modes_cms: t.List[t.ContextManager]

    def _get_class_list(
        self, cls_paths: t.Optional[t.Union[str, t.Sequence[str]]]
    ) -> list:
        from pipelime.choixe.utils.imports import import_symbol

        if not cls_paths:
            return []
        if isinstance(cls_paths, str):
            cls_paths = [cls_paths]
        if "_" in cls_paths:
            return []
        cls_paths = [c if "." in c else f"pipelime.items.{c}" for c in cls_paths]
        return [import_symbol(c) for c in cls_paths]

    def __init__(self, **data):
        import pipelime.items as pli

        super().__init__(**data)

        self._overridden_modes_cms = [
            pli.item_serialization_mode(m, *self._get_class_list(c))
            for m, c in self.override.items()
        ]

        self._disabled_modes_cms = [
            pli.item_disabled_serialization_modes(m, *self._get_class_list(c))
            for c, m in self.disable.items()
        ]

    def get_context_manager(self) -> t.ContextManager:
        from pipelime.utils.context_manager_list import ContextManagerList

        return ContextManagerList(
            *self._overridden_modes_cms, *self._disabled_modes_cms
        )


class OutputDatasetInterface(
    PydanticFieldNoDefaultMixin,
    pyd.BaseModel,
    extra="forbid",
    copy_on_model_validation="none",
):
    """Output dataset options.

    Examples:
        How to use it in your command::

            class EasyCommand(PipelimeCommand, title="easy"):
                input: InputDatasetInterface = InputDatasetInterface.pyd_field(
                    alias="i", piper_port=PiperPortType.INPUT
                )
                output: OutputDatasetInterface = OutputDatasetInterface.pyd_field(
                        alias="o", piper_port=PiperPortType.OUTPUT
                )
                grabber: GrabberInterface = GrabberInterface.pyd_field(
                    alias="g"
                )

                def run(self):
                    seq = self.input.create_reader()
                    seq = self.output.append_writer(seq)
                    self.grabber.grab_all(
                        seq,
                        grab_context_manager=self.output.serialization_cm(),
                        keep_order=False,
                        parent_cmd=self,
                        track_message=f"Executing ({len(seq)} samples)",
                    )
    """

    _default_type_description: t.ClassVar[t.Optional[str]] = "The output dataset."
    _compact_form: t.ClassVar[
        t.Optional[str]
    ] = "<folder>[,<exists_ok>[,<force_new_files>]]"

    folder: Path = pyd.Field(..., description="Dataset root folder.")
    zfill: t.Optional[int] = pyd.Field(None, description="Custom index zero-filling.")
    exists_ok: bool = pyd.Field(
        False, description="If False raises an error when `folder` exists."
    )
    serialization: SerializationModeInterface = pyd.Field(
        default_factory=SerializationModeInterface,
        description="Serialization modes for items and keys.",
    )
    schema_: t.Optional[SampleValidationInterface] = pyd.Field(
        None, alias="schema", description="Sample schema validation."
    )

    @pyd.validator("folder")
    def resolve_folder(cls, v: Path):
        # see https://bugs.python.org/issue38671
        return v.resolve().absolute()

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, value):
        if isinstance(value, OutputDatasetInterface):
            return value

        if isinstance(value, (str, bytes)):
            data = {}
            raw_data = str(value).split(",")
            data["folder"] = raw_data[0]
            if len(raw_data) > 1:
                # NB: the value is left as-is when it is not True nor False
                # to raise an error on validation
                data["exists_ok"] = (
                    True
                    if raw_data[1].lower() == "true"
                    else (False if raw_data[1].lower() == "false" else raw_data[1])
                )
            if len(raw_data) > 2:
                if raw_data[2].lower() == "true":
                    data["serialization"] = SerializationModeInterface(
                        override={"DEEP_COPY": None}
                    )
            value = data

        if isinstance(value, t.Mapping):
            return OutputDatasetInterface(**value)
        raise ValueError("Invalid output dataset definition.")

    def serialization_cm(self) -> t.ContextManager:
        return self.serialization.get_context_manager()

    def append_writer(self, sequence):
        writer = sequence.to_underfolder(
            folder=self.folder,
            zfill=self.zfill,
            exists_ok=self.exists_ok,
            key_serialization_mode=self.serialization.keys,
        )
        if self.schema_ is not None:
            writer = self.schema_.append_validator(writer)
        return writer

    def as_pipe(self):
        return {
            "to_underfolder": {
                "folder": self.folder,
                "zfill": self.zfill,
                "exists_ok": self.exists_ok,
                "key_serialization_mode": self.serialization.keys,
            },
            **({} if self.schema_ is None else self.schema_.as_pipe()),
        }

    def __repr__(self) -> str:
        return self.__piper_repr__()

    def __piper_repr__(self) -> str:
        return self.folder.as_posix()


class UrlDataModel(
    pyd.BaseModel,
    extra="forbid",
    copy_on_model_validation="none",
    underscore_attrs_are_private=True,
):
    """URL data model."""

    scheme: str = pyd.Field(..., description="The addressing scheme, eg, `s3`.")
    user: str = pyd.Field("", description="The user name.")
    password: str = pyd.Field("", description="The user password.")
    host: str = pyd.Field(..., description="The host name or ip address.")
    port: t.Optional[pyd.NonNegativeInt] = pyd.Field(
        None, description="The optional port number."
    )
    bucket: str = pyd.Field(..., description="The path to the remote data bucket.")
    args: t.Mapping[str, str] = pyd.Field(
        default_factory=dict, description="Optional remote-specific arguments."
    )

    def get_url(self):
        from pipelime.remotes import make_remote_url

        return make_remote_url(
            scheme=self.scheme,
            user=self.user,
            password=self.password,
            host=self.host,
            port=self.port,
            bucket=self.bucket,
            **self.args,
        )


class RemoteInterface(
    PydanticFieldNoDefaultMixin,
    pyd.BaseModel,
    extra="forbid",
    copy_on_model_validation="none",
    underscore_attrs_are_private=True,
):
    """Remote data lake options."""

    _default_type_description: t.ClassVar[
        t.Optional[str]
    ] = "Remote data lakes addresses."
    _compact_form: t.ClassVar[t.Optional[str]] = "<url>"

    url: t.Union[str, UrlDataModel] = pyd.Field(
        ...,
        description=(
            "The remote data lake URL. You can user the format "
            "`s3://user:password@host:port/bucket?kw1=arg1:kw2=arg2`."
        ),
    )

    _parsed_url: ParseResult

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, value):
        if isinstance(value, RemoteInterface):
            return value
        if isinstance(value, (str, bytes)):
            return RemoteInterface(url=str(value))
        if isinstance(value, t.Mapping):
            return RemoteInterface(**value)
        raise ValueError("Invalid remote definition.")

    def __init__(self, **data):
        from urllib.parse import urlparse

        super().__init__(**data)
        self._parsed_url = (
            self.url.get_url()
            if isinstance(self.url, UrlDataModel)
            else urlparse(self.url)
        )

    def get_url(self):
        return self._parsed_url


class ToyDatasetInterface(
    pyd.BaseModel, extra="forbid", copy_on_model_validation="none"
):
    """Toy dataset creation options."""

    length: int = pyd.Field(..., description="The number of samples to generate.")
    with_images: bool = pyd.Field(True, description="Whether to generate images.")
    with_masks: bool = pyd.Field(
        True, description="Whether to generate masks with object labels."
    )
    with_instances: bool = pyd.Field(
        True, description="Whether to generate images with object indexes."
    )
    with_objects: bool = pyd.Field(True, description="Whether to generate objects.")
    with_bboxes: bool = pyd.Field(
        True, description="Whether to generate objects' bboxes."
    )
    with_kpts: bool = pyd.Field(
        True, description="Whether to generate objects' keypoints."
    )
    image_size: int = pyd.Field(256, description="The size of the generated images.")
    key_format: str = pyd.Field(
        "*",
        description=(
            "The sample key format. Any `*` will be replaced with the "
            "base key name, eg, `my_*_key` on [`image`, `mask`] generates "
            "`my_image_key` and `my_mask_key`. If no `*` is found, the string is "
            "suffixed to the base key name, ie, `MyKey` on `image` gives "
            "`imageMyKey`. If empty, the base key name will be used."
        ),
    )
    max_labels: int = pyd.Field(
        5, description="The maximum number of object labels in the dataset."
    )
    objects_range: t.Tuple[int, int] = pyd.Field(
        (1, 5), description="The (min, max) number of objects in each sample."
    )

    @pyd.validator("key_format")
    def validate_key_format(cls, v):
        if "*" in v:
            return v
        return "*" + v

    def create_dataset_generator(self):
        from pipelime.sequences import SamplesSequence

        return SamplesSequence.toy_dataset(  # type: ignore
            length=self.length,
            with_images=self.with_images,
            with_masks=self.with_masks,
            with_instances=self.with_instances,
            with_objects=self.with_objects,
            with_bboxes=self.with_bboxes,
            with_kpts=self.with_kpts,
            image_size=self.image_size,
            key_format=self.key_format,
            max_labels=self.max_labels,
            objects_range=self.objects_range,
        )



import hashlib
import typing as t

from loguru import logger

from pipelime.remotes.base import BaseRemote, NetlocData


class S3Remote(BaseRemote):
    _HASH_FN_KEY_ = "__HASH_FN__"
    _DEFAULT_HASH_FN_ = "sha256"

    def __init__(self, netloc_data: NetlocData):
        """S3-compatible remote. Credential can be passed or retrieved from:
            * env var:
                ** access key: AWS_ACCESS_KEY_ID, AWS_ACCESS_KEY, MINIO_ACCESS_KEY
                ** secret key: AWS_SECRET_ACCESS_KEY, AWS_SECRET_KEY, MINIO_SECRET_KEY
                ** session token: AWS_SESSION_TOKEN
            * config files:
                ** ~/.aws/credentials
                ** ~/[.]mc/config.json

        Optional init arguments:
            * session (str): the session token (default: None)
            * secure (bool): establish a secure connection (default: True)
            * region (str): the preferred region (default: None)

        :param netloc_data: the network data info.
        :type netloc: NetlocData
        """
        super().__init__(netloc_data)

        try:
            from minio import Minio
            from minio.credentials import (
                AWSConfigProvider,
                ChainedProvider,
                EnvAWSProvider,
                EnvMinioProvider,
                MinioClientConfigProvider,
            )

            self._client = Minio(
                endpoint=netloc_data.host,
                access_key=netloc_data.user,
                secret_key=netloc_data.password,
                session_token=netloc_data.init_args.get("session", None),
                secure=netloc_data.init_args.get("secure", True),  # type: ignore
                region=netloc_data.init_args.get("region", None),
                credentials=ChainedProvider(
                    [
                        # AWS_ACCESS_KEY_ID|AWS_ACCESS_KEY,
                        # AWS_SECRET_ACCESS_KEY|AWS_SECRET_KEY,
                        # AWS_SESSION_TOKEN
                        EnvAWSProvider(),
                        # MINIO_ACCESS_KEY, MINIO_SECRET_KEY
                        EnvMinioProvider(),
                        # ~/.aws/credentials
                        AWSConfigProvider(),
                        # ~/[.]mc/config.json
                        MinioClientConfigProvider(),
                    ]
                ),
            )
        except ModuleNotFoundError:  # pragma: no cover
            logger.error("S3 remote needs `minio` python package.")
            self._client = None

    def _maybe_create_bucket(self, target_base_path: str):
        if not self._client.bucket_exists(target_base_path):  # type: ignore
            logger.info(
                f"Creating bucket '{target_base_path}' on S3 remote {self.netloc}."
            )
            self._client.make_bucket(target_base_path)  # type: ignore

    def _get_hash_fn(self, target_base_path: str) -> t.Any:
        if self.is_valid:
            try:
                self._maybe_create_bucket(target_base_path)
                tags = self._client.get_bucket_tags(target_base_path)  # type: ignore
                if tags is None:
                    from minio.commonconfig import Tags

                    tags = Tags()
                hash_fn_name = tags.get(self._HASH_FN_KEY_)

                # try-get
                if isinstance(hash_fn_name, str) and len(hash_fn_name) > 0:
                    try:
                        hash_fn = getattr(hashlib, hash_fn_name)
                        return hash_fn()
                    except AttributeError:
                        pass

                tags[self._HASH_FN_KEY_] = self._DEFAULT_HASH_FN_
                self._client.set_bucket_tags(target_base_path, tags)  # type: ignore

                hash_fn = getattr(hashlib, self._DEFAULT_HASH_FN_)
                return hash_fn()
            except Exception as exc:
                logger.debug(str(exc))
        return None

    def target_exists(self, target_base_path: str, target_name: str) -> bool:
        if self.is_valid:
            try:
                objlist = self._client.list_objects(  # type: ignore
                    target_base_path, prefix=target_name
                )
                _ = next(objlist)
                return True
            except StopIteration:
                return False
            except Exception as exc:
                logger.debug(str(exc))
        return False

    def _upload(
        self,
        local_stream: t.BinaryIO,
        local_stream_size: int,
        target_base_path: str,
        target_name: str,
    ) -> bool:
        if self.is_valid:
            try:
                self._maybe_create_bucket(target_base_path)
                self._client.put_object(  # type: ignore
                    bucket_name=target_base_path,
                    object_name=target_name,
                    data=local_stream,
                    length=local_stream_size,
                )
                return True
            except Exception as exc:
                logger.debug(str(exc))
                return False

        return False

    def _download(
        self,
        local_stream: t.BinaryIO,
        source_base_path: str,
        source_name: str,
        source_offset: int,
    ) -> bool:
        if self.is_valid:
            if not self._client.bucket_exists(source_base_path):  # type: ignore
                logger.debug(
                    f"Bucket '{source_base_path}' does not exist "
                    f"on S3 remote '{self.netloc}'."
                )
                return False

            response = None
            ok = False
            try:
                response = self._client.get_object(  # type: ignore
                    bucket_name=source_base_path,
                    object_name=source_name,
                    offset=source_offset,
                )
                for data in response.stream(amt=1024 * 1024):
                    local_stream.write(data)
                ok = True
            except Exception as exc:
                logger.debug(str(exc))
                return False
            finally:
                if response:
                    response.close()
                    response.release_conn()
                return ok

        return False

    @classmethod
    def scheme(cls) -> str:
        return "s3"

    @property
    def is_valid(self) -> bool:
        return self._client is not None



import typing as t

from pipelime.items import Item


class BinaryItem(Item[bytes]):
    @classmethod
    def file_extensions(cls) -> t.Sequence[str]:
        return (".bin",)

    @classmethod
    def decode(cls, fp: t.BinaryIO) -> bytes:
        return fp.read()

    @classmethod
    def encode(cls, value: bytes, fp: t.BinaryIO):
        fp.write(value)

    @classmethod
    def validate(cls, raw_data: t.Any) -> bytes:
        if not isinstance(raw_data, bytes):
            raise ValueError(f"{cls}: raw data must be bytes.")  # pragma: no cover
        return raw_data

    @classmethod
    def pl_pretty_data(cls, value: bytes) -> t.Any:
        from rich.text import Text

        return Text(value.hex(), overflow="ellipsis") + Text(f" ({len(value)} bytes)")



import pytest

from pipelime.sequences import SamplesSequence


class TestValidation:
    def _schema_check(self, dataset, schema_def, lazy, should_fail):
        try:
            seq = SamplesSequence.from_underfolder(  # type: ignore
                dataset
            ).validate_samples(sample_schema=schema_def, lazy=lazy)
            try:
                for _ in seq:
                    pass
                assert not should_fail
            except ValueError:
                assert should_fail and lazy
        except ValueError:
            assert should_fail and not lazy

    @pytest.mark.parametrize("lazy", [True, False])
    @pytest.mark.parametrize("ignore_extra_keys", [True, False])
    def test_validate_full_schema(
        self, minimnist_dataset: dict, lazy, ignore_extra_keys
    ):
        from typing import Optional

        import pydantic as pyd

        import pipelime.items as pli

        class MySchema(
            pyd.BaseModel,
            arbitrary_types_allowed=True,
            extra=pyd.Extra.ignore if ignore_extra_keys else pyd.Extra.forbid,
        ):
            cfg: pli.YamlMetadataItem
            numbers: pli.TxtNumpyItem
            pose: pli.TxtNumpyItem
            image: pli.PngImageItem
            label: pli.TxtNumpyItem
            mask: pli.PngImageItem
            metadata: pli.JsonMetadataItem
            points: pli.TxtNumpyItem
            invalid_key: Optional[pli.TxtNumpyItem] = None

        self._schema_check(
            minimnist_dataset["path"], MySchema, lazy=lazy, should_fail=False
        )

    @pytest.mark.parametrize("lazy", [True, False])
    @pytest.mark.parametrize("ignore_extra_keys", [True, False])
    def test_validate_partial_schema(
        self, minimnist_dataset: dict, lazy, ignore_extra_keys
    ):
        from typing import Optional

        import pydantic as pyd

        import pipelime.items as pli

        class MySchema(
            pyd.BaseModel,
            arbitrary_types_allowed=True,
            extra="ignore" if ignore_extra_keys else "forbid",
        ):
            cfg: pli.YamlMetadataItem
            numbers: pli.TxtNumpyItem
            pose: pli.TxtNumpyItem
            invalid_key: Optional[pli.TxtNumpyItem] = None

        self._schema_check(
            minimnist_dataset["path"],
            MySchema,
            lazy=lazy,
            should_fail=not ignore_extra_keys,
        )

    @pytest.mark.parametrize("lazy", [True, False])
    @pytest.mark.parametrize("ignore_extra_keys", [True, False])
    def test_validate_schema_with_base_classes(
        self, minimnist_dataset: dict, lazy, ignore_extra_keys
    ):
        from typing import Optional

        import pydantic as pyd

        import pipelime.items as pli

        class MySchema(
            pyd.BaseModel,
            arbitrary_types_allowed=True,
            extra=pyd.Extra.ignore if ignore_extra_keys else pyd.Extra.forbid,
        ):
            cfg: pli.MetadataItem
            numbers: pli.NumpyItem
            pose: pli.TxtNumpyItem
            image: pli.NumpyItem
            label: pli.Item
            mask: pli.ImageItem
            invalid_key: Optional[pli.Item] = None

        self._schema_check(
            minimnist_dataset["path"],
            MySchema,
            lazy=lazy,
            should_fail=not ignore_extra_keys,
        )



import time
from typing import List, Optional

from pipelime.piper.progress.listener.base import (
    Listener,
    ListenerCallback,
    ProgressReceiver,
)
from pipelime.piper.progress.model import OperationInfo, ProgressUpdate


class MockProgressReceiver(ProgressReceiver):
    def __init__(self, token: str, op_info: OperationInfo) -> None:
        super().__init__(token)
        self._op_info = op_info
        self._progress = 0

    def receive(self) -> Optional[ProgressUpdate]:
        if self._progress < self._op_info.total:
            self._progress += 1
        time.sleep(0.01)
        return ProgressUpdate(op_info=self._op_info, progress=self._progress)


class MockListenerCallback(ListenerCallback):
    def __init__(self) -> None:
        super().__init__()
        self.on_start_called = 0
        self.on_update_called: List[ProgressUpdate] = []
        self.on_stop_called = 0

    def on_start(self) -> None:
        self.on_start_called += 1

    def on_update(self, prog: ProgressUpdate) -> None:
        self.on_update_called.append(prog)

    def on_stop(self) -> None:
        self.on_stop_called += 1


class TestListener:
    def test_listener(self):
        # Create listener
        op_info = OperationInfo(
            token="token", node="node", chunk=1, message="message", total=10
        )
        receiver = MockProgressReceiver("token", op_info)
        callback = MockListenerCallback()
        listener = Listener(receiver, callback)

        # Start listener
        listener.start()

        # Wait some time
        time.sleep(0.2)

        # Stop listener
        listener.stop()

        # Assert that listener was called correctly
        assert callback.on_start_called == 1
        assert len(callback.on_update_called) > 0
        assert callback.on_stop_called == 1

        assert callback.on_update_called[-1].op_info == op_info
        assert callback.on_update_called[-1].progress > 0



import time

import zmq

from pipelime.piper.progress.model import ProgressUpdate
from pipelime.piper.progress.tracker.base import TrackCallback


class ZmqTrackCallback(TrackCallback):
    """ZMQ tracker callback"""

    def __init__(self, addr: str = "tcp://*:5556") -> None:
        super().__init__()
        self._addr = addr
        self._socket = None

    def _send(self, prog: ProgressUpdate) -> None:
        topic = prog.op_info.token
        if self._socket is not None:
            self._socket.send_multipart([topic.encode(), prog.json().encode()])

    def on_start(self, prog: ProgressUpdate) -> None:
        self._socket = zmq.Context().socket(zmq.PUB)
        self._socket.bind(self._addr)

        # Wait for the socket to be ready...
        # Apparently, this is the only way to do it. I don't know why.
        time.sleep(1)

        self._send(prog)

    def on_advance(self, prog: ProgressUpdate) -> None:
        self._send(prog)

    def on_finish(self, prog: ProgressUpdate) -> None:
        self._send(prog)

        if self._socket is not None:
            self._socket.close()
